import marshal
import functools
from . import analyzeSBML
from collections import Counter, defaultdict
import itertools
from copy import deepcopy, copy
from bionetgen.atomizer.utils.util import logMess, memoize, memoizeMapped
from . import atomizationAux as atoAux
import bionetgen.atomizer.utils.pathwaycommons as pwcm


class SCTSolver:
    def __init__(self, database, memoizedResolver=False):
        self.database = database
        self.memoizedResolver = memoizedResolver
        self.graph_map = {}
        self.dg = None

    def createSpeciesCompositionGraph(
        self,
        parser,
        configurationFile,
        namingConventions,
        speciesEquivalences=None,
        bioGridFlag=False,
    ):
        """
        Main method for the SCT creation.

        It first does stoichiometry analysis, then lexical...
        """

        _, rules, _ = parser.getReactions(atomize=True, database=self.database)
        molecules, _, _, _, _, _ = parser.getSpecies()
        self.database.sbmlAnalyzer = analyzeSBML.SBMLAnalyzer(
            parser,
            configurationFile,
            namingConventions,
            speciesEquivalences,
            conservationOfMass=True,
        )

        # classify reactions
        (
            self.database.classifications,
            equivalenceTranslator,
            self.database.eequivalenceTranslator,
            indirectEquivalenceTranslator,
            adhocLabelDictionary,
            lexicalDependencyGraph,
            userEquivalenceTranslator,
        ) = self.database.sbmlAnalyzer.classifyReactions(rules, molecules, {})
        self.database.reactionProperties = (
            self.database.sbmlAnalyzer.getReactionProperties()
        )

        syndecs = [
            1 if i == "Generation" or i == "Decay" else 0
            for i in self.database.classifications
        ]
        # user defined and lexical analysis naming conventions are stored here
        self.database.reactionProperties.update(adhocLabelDictionary)

        (
            self.database.translator,
            self.database.userLabelDictionary,
            self.database.lexicalLabelDictionary,
            self.database.partialUserLabelDictionary,
        ) = self.database.sbmlAnalyzer.getUserDefinedComplexes()
        self.database.dependencyGraph = {}
        self.database.alternativeDependencyGraph = {}
        # fill in the annotation dictionary
        self.database.annotationDict = parser.getFullAnnotation()
        # just molecule names without parenthesis
        strippedMolecules = [x.strip("()") for x in molecules]
        # self.database.annotationDict = {}

        # ###dependency graph
        # binding reactions
        for reaction, classification in zip(rules, self.database.classifications):
            self.bindingReactionsAnalysis(
                self.database.dependencyGraph,
                list(atoAux.parseReactions(reaction)),
                classification,
            )
        # lexical dependency graph contains lexically induced binding compositions. atomizer gives preference to binding obtained this way as opposed to stoichiometry
        # stronger bounds on stoichiometry based binding can be defined in
        # reactionDefinitions.json.

        for element in lexicalDependencyGraph:

            if (
                element in self.database.dependencyGraph
                and element not in self.database.userLabelDictionary
            ):
                if len(lexicalDependencyGraph[element]) == 0:
                    continue
                """
                oldDependency = self.database.dependencyGraph[element]
                
                if sorted(lexicalDependencyGraph[element][0]) in [sorted(x) for x in oldDependency]:
                    # if len(oldDependency) > 1:
                    #    logMess('DEBUG:Atomization', 'Species {0} was confirmed to be {1} based on lexical information'.format(element,lexicalDependencyGraph[element]))
                    self.database.dependencyGraph[
                        element] = lexicalDependencyGraph[element]
                else:
                    # logMess('INFO:Atomization', 'Species {0} was determined to be {1} instead of {2} based on \
                    # lexical information'.format(element,
                    # lexicalDependencyGraph[element], oldDependency))
                """
                if self.database.dependencyGraph[element] != []:
                    self.database.alternativeDependencyGraph[
                        element
                    ] = lexicalDependencyGraph[element]
                else:
                    logMess(
                        "INFO:LAE009",
                        "{0}: being set to be a modification of constructed species {1}".format(
                            element, lexicalDependencyGraph[element][0]
                        ),
                    )
                    atoAux.addToDependencyGraph(
                        self.database.dependencyGraph,
                        element,
                        lexicalDependencyGraph[element][0],
                    )
            else:
                if element not in strippedMolecules:
                    self.database.constructedSpecies.add(element)
                self.database.dependencyGraph[element] = lexicalDependencyGraph[element]
            # Check if I'm using a molecule that hasn't been used yet
            for dependencyCandidate in self.database.dependencyGraph[element]:
                for molecule in [
                    x
                    for x in dependencyCandidate
                    if x not in self.database.dependencyGraph
                ]:
                    # this is a species that was not originally in the model. in case theres conflict later this is
                    # to indicate it is given less priority
                    self.database.dependencyGraph[molecule] = []

        # user defined transformations
        for key in userEquivalenceTranslator:
            for namingEquivalence in userEquivalenceTranslator[key]:
                baseElement = min(namingEquivalence, key=len)
                modElement = max(namingEquivalence, key=len)
                if baseElement not in self.database.dependencyGraph:
                    self.database.dependencyGraph[baseElement] = []
                atoAux.addToDependencyGraph(
                    self.database.dependencyGraph, modElement, [baseElement]
                )

        # self.database.eequivalence translator contains 1:1 equivalences
        # FIXME: do we need this update step or is it enough with the later one?
        # catalysis reactions
        """
        for key in self.database.eequivalenceTranslator:
            for namingEquivalence in self.database.eequivalenceTranslator[key]:
                baseElement = min(namingEquivalence, key=len)
                modElement = max(namingEquivalence, key=len)
                if key != 'Binding':
                    if baseElement not in self.database.dependencyGraph or self.database.dependencyGraph[baseElement] == []:
                        if modElement not in self.database.dependencyGraph or self.database.dependencyGraph[modElement] == []:
                            self.database.dependencyGraph[baseElement] = []
                        # do we have a meaningful reverse dependence?
                        # elif all([baseElement not in x for x in self.database.dependencyGraph[modElement]]):
                        #    atoAux.addToDependencyGraph(self.database.dependencyGraph,baseElement,[modElement])
                        #    continue

                            if baseElement in self.database.annotationDict and modElement in self.database.annotationDict:
                                baseSet = set([y for x in self.database.annotationDict[
                                              baseElement] for y in self.database.annotationDict[baseElement][x]])
                                modSet = set([y for x in self.database.annotationDict[
                                             modElement] for y in self.database.annotationDict[modElement][x]])
                                if len(baseSet.intersection(modSet)) > 0 or len(baseSet) == 0 or len(modSet) == 0:
                                    atoAux.addToDependencyGraph(self.database.dependencyGraph, modElement,
                                                         [baseElement])
                                else:
                                    logMess("ERROR:ANN201", "{0} and {1} have a direct correspondence according to reaction information however their annotations are completely different.".format(
                                        baseElement, modElement))
                            else:
                                atoAux.addToDependencyGraph(self.database.dependencyGraph, modElement,
                                                     [baseElement])
        """
        # include user label information.
        for element in self.database.userLabelDictionary:
            if self.database.userLabelDictionary[element] in [0, [(0,)]]:
                self.database.dependencyGraph[element] = ["0"]
            elif (
                len(self.database.userLabelDictionary[element][0]) == 0
                or element == self.database.userLabelDictionary[element][0][0]
            ):

                self.database.dependencyGraph[element] = []
            else:
                self.database.dependencyGraph[element] = [
                    list(self.database.userLabelDictionary[element][0])
                ]
                # If the user is introducing a new molecule term, add it to the SCT
                if (
                    self.database.userLabelDictionary[element][0][0]
                    not in self.database.dependencyGraph
                ):
                    self.database.dependencyGraph[
                        self.database.userLabelDictionary[element][0][0]
                    ] = []

        # add species elements defined by the user into the naming convention
        # definition
        molecules.extend(
            [
                "{0}()".format(x)
                for x in self.database.userLabelDictionary
                if "{0}()".format(x) not in molecules
            ]
        )
        # recalculate 1:1 equivalences now with binding information
        (
            _,
            _,
            self.database.eequivalenceTranslator2,
            _,
            adhocLabelDictionary,
            _,
            _,
        ) = self.database.sbmlAnalyzer.classifyReactions(
            rules, molecules, self.database.dependencyGraph
        )
        self.database.reactionProperties.update(adhocLabelDictionary)
        # update catalysis equivalences
        # catalysis reactions
        for key in self.database.eequivalenceTranslator2:
            for namingEquivalence in self.database.eequivalenceTranslator2[key]:

                baseElement = min(namingEquivalence, key=len)
                modElement = max(namingEquivalence, key=len)
                # dont overwrite user information
                if (
                    key != "Binding"
                    and modElement not in self.database.userLabelDictionary
                ):
                    if baseElement not in self.database.dependencyGraph:
                        self.database.constructedSpecies.add(baseElement)
                        self.database.dependencyGraph[baseElement] = []
                    if modElement not in self.database.dependencyGraph or not [
                        True
                        for x in self.database.dependencyGraph[modElement]
                        if baseElement in x and len(x) > 1
                    ]:
                        if (
                            baseElement in self.database.annotationDict
                            and modElement in self.database.annotationDict
                        ):
                            baseSet = set(
                                [
                                    y
                                    for x in self.database.annotationDict[baseElement]
                                    for y in self.database.annotationDict[baseElement][
                                        x
                                    ]
                                ]
                            )
                            modSet = set(
                                [
                                    y
                                    for x in self.database.annotationDict[modElement]
                                    for y in self.database.annotationDict[modElement][x]
                                ]
                            )
                            if (
                                len(baseSet.intersection(modSet)) > 0
                                or len(baseSet) == 0
                                or len(modSet) == 0
                            ):
                                if modElement not in self.database.dependencyGraph:
                                    # if the entry doesnt exist from previous information accept this
                                    atoAux.addToDependencyGraph(
                                        self.database.dependencyGraph,
                                        modElement,
                                        [baseElement],
                                    )
                                else:
                                    # otherwise add it to the lexical repository
                                    atoAux.addToDependencyGraph(
                                        self.database.alternativeDependencyGraph,
                                        modElement,
                                        [baseElement],
                                    )
                            else:
                                baseDB = set(
                                    [
                                        x.split("/")[-2]
                                        for x in baseSet
                                        if "identifiers.org" in x
                                    ]
                                )
                                modDB = set(
                                    [
                                        x.split("/")[-2]
                                        for x in modSet
                                        if "identifiers.org" in x
                                    ]
                                )
                                # it is still ok if they each refer to different self.databases
                                if len(baseDB.intersection(modDB)) == 0:
                                    if modElement not in self.database.dependencyGraph:
                                        # if the entry doesnt exist from previous information accept this
                                        atoAux.addToDependencyGraph(
                                            self.database.dependencyGraph,
                                            modElement,
                                            [baseElement],
                                        )
                                    else:
                                        # otherwise add it to the lexical repository
                                        atoAux.addToDependencyGraph(
                                            self.database.alternativeDependencyGraph,
                                            modElement,
                                            [baseElement],
                                        )
                                else:
                                    logMess(
                                        "WARNING:ANN201",
                                        "{0} and {1} have a direct correspondence according to reaction information however their annotations are completely different.".format(
                                            baseElement, modElement
                                        ),
                                    )
                        else:
                            atoAux.addToDependencyGraph(
                                self.database.dependencyGraph, modElement, [baseElement]
                            )
                    else:
                        logMess(
                            "WARNING:ATO114",
                            "Definition conflict between binding information {0} and lexical analyis {1} for molecule {2},\
    choosing binding".format(
                                self.database.dependencyGraph[modElement],
                                baseElement,
                                modElement,
                            ),
                        )

        # non lexical-analysis catalysis reactions
        if self.database.forceModificationFlag:
            for reaction, classification in zip(rules, self.database.classifications):
                preaction = list(atoAux.parseReactions(reaction))
                if len(preaction[0]) == 1 and len(preaction[1]) == 1:
                    if (preaction[0][0] in [0, "0"]) or (preaction[1][0] in [0, "0"]):
                        continue
                    if preaction[1][0].lower() in preaction[0][0].lower() or len(
                        preaction[1][0]
                    ) < len(preaction[0][0]):
                        base = preaction[1][0]
                        mod = preaction[0][0]
                    else:
                        mod = preaction[1][0]
                        base = preaction[0][0]
                    if (
                        self.database.dependencyGraph[mod] == []
                        and mod not in self.database.userLabelDictionary
                    ):

                        if (
                            base in self.database.userLabelDictionary
                            and self.database.userLabelDictionary[base] == 0
                        ):
                            continue
                        if (
                            mod in self.database.userLabelDictionary
                            and self.database.userLabelDictionary[mod] == 0
                        ):
                            continue
                        if [mod] in self.database.dependencyGraph[base]:
                            continue

                        # can we just match it up through existing species instead of forcing a modification?
                        greedyMatch = (
                            self.database.sbmlAnalyzer.greedyModificationMatching(
                                mod, self.database.dependencyGraph.keys()
                            )
                        )

                        if greedyMatch not in [-1, -2, []]:
                            self.database.dependencyGraph[mod] = [greedyMatch]
                            if mod in self.database.alternativeDependencyGraph:
                                del self.database.alternativeDependencyGraph[mod]
                            logMess(
                                "INFO:LAE006",
                                "{0}: Mapped to {1} using lexical analysis/greedy matching".format(
                                    mod, greedyMatch
                                ),
                            )
                            continue

                        # if the annotations have no overlap whatsoever don't force
                        # this modifications
                        if (
                            base in self.database.annotationDict
                            and mod in self.database.annotationDict
                        ):
                            baseSet = set(
                                [
                                    y
                                    for x in self.database.annotationDict[base]
                                    for y in self.database.annotationDict[base][x]
                                ]
                            )
                            modSet = set(
                                [
                                    y
                                    for x in self.database.annotationDict[mod]
                                    for y in self.database.annotationDict[mod][x]
                                ]
                            )
                            if (
                                (len(baseSet.intersection(modSet))) == 0
                                and len(baseSet) > 0
                                and len(modSet) > 0
                            ):
                                baseDB = set(
                                    [
                                        x.split("/")[-2]
                                        for x in baseSet
                                        if "identifiers.org" in x
                                    ]
                                )
                                modDB = set(
                                    [
                                        x.split("/")[-2]
                                        for x in modSet
                                        if "identifiers.org" in x
                                    ]
                                )
                                # we stil ahve to check that they both reference the same self.database
                                if len(baseDB.intersection(modDB)) > 0:
                                    logMess(
                                        "WARNING:ANN201",
                                        "{0} and {1} have a direct correspondence according to reaction information however their annotations are completely different.".format(
                                            base, mod
                                        ),
                                    )
                                    continue
                        self.database.dependencyGraph[mod] = [[base]]

        """
        #complex catalysis reactions
        for key in indirectEquivalenceTranslator:
            #first remove these entries from the dependencyGraph since
            #they are not true bindingReactions
            for namingEquivalence in indirectEquivalenceTranslator[key]:
                removedElement = ''
                tmp3 = deepcopy(namingEquivalence[1])
                if tmp3 in self.database.dependencyGraph[namingEquivalence[0][0]]:
                    removedElement = namingEquivalence[0][0]
                elif tmp3 in self.database.dependencyGraph[namingEquivalence[0][1]]:
                    removedElement = namingEquivalence[0][1]

                else:
                    tmp3.reverse()
                    if tmp3 in self.database.dependencyGraph[namingEquivalence[0][0]]:
                        removedElement = namingEquivalence[0][0]

                    elif tmp3 in self.database.dependencyGraph[namingEquivalence[0][1]]:
                        removedElement = namingEquivalence[0][1]


                #then add the new, true dependencies
                #if its not supposed to be a basic element
                tmp = [x for x in namingEquivalence[1] if x not in namingEquivalence[2]]
                tmp.extend([x for x in namingEquivalence[2] if x not in namingEquivalence[1]])
                tmp2 = deepcopy(tmp)
                tmp2.reverse()
                ##TODO: map back for the elements in namingEquivalence[2]
                if tmp not in self.database.dependencyGraph[namingEquivalence[3][0]] \
                    and tmp2 not in self.database.dependencyGraph[namingEquivalence[3][0]]:
                    if sorted(tmp) == sorted(tmp3):
                        continue
                    if all(x in self.database.dependencyGraph for x in tmp):
                        if removedElement in self.database.dependencyGraph:
                            self.database.dependencyGraph[removedElement].remove(tmp3)
                        logMess('INFO:Atomization','Removing {0}={1} and adding {2}={3} instead\
     from the dependency list since we determined it is not a true binding reaction based on lexical analysis'\
                        .format(removedElement,tmp3,namingEquivalence[3][0],tmp))
                        self.database.dependencyGraph[namingEquivalence[3][0]] = [tmp]
                    else:
                        logMess('WARNING:Atomization','We determined that {0}={1} based on lexical analysis instead of \
    {2}={3} (stoichiometry) but one of the constituent components in {1} is not a molecule so no action was taken'.format(namingEquivalence[3][0],
    tmp,removedElement,tmp3))
        #user defined stuff
    """

        # stuff obtained from string similarity analysis
        for element in self.database.lexicalLabelDictionary:
            # similarity analysis has less priority than anything we discovered
            # before
            if (
                element in self.database.dependencyGraph
                and len(self.database.dependencyGraph[element]) > 0
            ):
                continue

            if (
                len(self.database.lexicalLabelDictionary[element][0]) == 0
                or element == self.database.lexicalLabelDictionary[element][0][0]
            ):
                self.database.constructedSpecies.add(element)
                atoAux.addToDependencyGraph(self.database.dependencyGraph, element, [])
            else:
                # logMess('INFO:Atomization', ' added induced speciesStructure {0}={1}'
                #         .format(element, self.database.lexicalLabelDictionary[element][0]))
                self.database.dependencyGraph[element] = [
                    list(self.database.lexicalLabelDictionary[element][0])
                ]

        # Now let's go for annotation analysis and last resort stuff on the remaining orphaned molecules
        orphanedSpecies = [
            x
            for x in strippedMolecules
            if x not in self.database.dependencyGraph
            or self.database.dependencyGraph[x] == []
        ]
        orphanedSpecies.extend(
            [
                x
                for x in self.database.dependencyGraph
                if self.database.dependencyGraph[x] == [] and x not in orphanedSpecies
            ]
        )

        # Fill SCT with annotations for those species that still dont have any
        # mapping

        annotationDependencyGraph, _ = self.fillSCTwithAnnotationInformation(
            orphanedSpecies, self.database.annotationDict
        )

        # use an empty dictionary if we wish to turn off annotation information in atomization
        # annotationDependencyGraph = {}

        for annotatedSpecies in annotationDependencyGraph:
            if (
                len(annotationDependencyGraph[annotatedSpecies]) > 0
                and annotatedSpecies not in self.database.userLabelDictionary
            ):
                atoAux.addToDependencyGraph(
                    self.database.dependencyGraph,
                    annotatedSpecies,
                    annotationDependencyGraph[annotatedSpecies][0],
                )
                logMess(
                    "INFO:ANN004",
                    "Added equivalence from annotation information {0}={1}".format(
                        annotatedSpecies, annotationDependencyGraph[annotatedSpecies][0]
                    ),
                )
                for element in annotationDependencyGraph[annotatedSpecies][0]:
                    # in case one of the compositional elements is not yet in the
                    # dependency graph
                    if element not in self.database.dependencyGraph:
                        atoAux.addToDependencyGraph(
                            self.database.dependencyGraph, element, []
                        )

        # can we now add information to the non orphaned species? maybe annotation tells me stuff that contradicts the reaction-network
        nonOrphanedSpecies = [x for x in strippedMolecules if x not in orphanedSpecies]

        annotationDependencyGraph, _ = self.fillSCTwithAnnotationInformation(
            nonOrphanedSpecies,
            self.database.annotationDict,
            self.database,
            tentativeFlag=True,
        )

        orphanedSpecies = [
            x
            for x in strippedMolecules
            if x not in self.database.dependencyGraph
            or self.database.dependencyGraph[x] == []
        ]
        orphanedSpecies.extend(
            [
                x
                for x in self.database.dependencyGraph
                if self.database.dependencyGraph[x] == [] and x not in orphanedSpecies
            ]
        )

        orphanedSpecies.extend(self.database.constructedSpecies)
        strippedMolecules.extend(
            [x.strip("()") for x in self.database.constructedSpecies]
        )
        # TODO: merge both lists and use them as a tiebreaker for consolidation
        # completeAnnotationDependencyGraph, completePartialMatches = fillSCTwithAnnotationInformation(strippedMolecules, annotationDict, self.database, False)
        # pure lexical analysis for the remaining orphaned molecules
        (
            tmpDependency,
            self.database.tmpEquivalence,
        ) = self.database.sbmlAnalyzer.findClosestModification(
            orphanedSpecies,
            strippedMolecules,
            self.database.annotationDict,
            self.database.dependencyGraph,
        )

        for species in tmpDependency:
            if species not in self.database.userLabelDictionary:
                if tmpDependency[species] == []:
                    atoAux.addToDependencyGraph(
                        self.database.dependencyGraph, species, []
                    )
                for instance in tmpDependency[species]:
                    atoAux.addToDependencyGraph(
                        self.database.dependencyGraph, species, instance
                    )
                    if (
                        len(instance) == 1
                        and instance[0] not in self.database.dependencyGraph
                    ):
                        atoAux.addToDependencyGraph(
                            self.database.dependencyGraph, instance[0], []
                        )

        orphanedSpecies = [
            x
            for x in strippedMolecules
            if x not in self.database.dependencyGraph
            or self.database.dependencyGraph[x] == []
        ]

        orphanedSpecies.extend(
            [
                x
                for x in self.database.dependencyGraph
                if self.database.dependencyGraph[x] == [] and x not in orphanedSpecies
            ]
        )

        orphanedSpecies.extend(self.database.constructedSpecies)
        # greedy lexical analysis for the remaining orhpaned species
        for reactant in orphanedSpecies:
            greedyMatch = self.database.sbmlAnalyzer.greedyModificationMatching(
                reactant, self.database.dependencyGraph.keys()
            )
            if greedyMatch not in [-1, -2, []]:
                atoAux.addToDependencyGraph(
                    self.database.dependencyGraph, reactant, greedyMatch
                )
                logMess(
                    "INFO:LAE006",
                    "Mapped {0} to {1} using lexical analysis/greedy matching".format(
                        reactant, greedyMatch
                    ),
                )
        if len(self.database.constructedSpecies) > 0:
            logMess(
                "INFO:SCT031",
                "The following species names do not appear in the original model but where created to have more appropiate naming conventions: [{0}]".format(
                    ",".join(self.database.constructedSpecies)
                ),
            )

        # initialize and remove zero elements
        (
            self.database.prunnedDependencyGraph,
            self.database.weights,
            unevenElementDict,
            self.database.artificialEquivalenceTranslator,
        ) = self.consolidateDependencyGraph(
            self.database.dependencyGraph,
            equivalenceTranslator,
            self.database.eequivalenceTranslator,
            self.database.sbmlAnalyzer,
        )
        return self.database

    def bindingReactionsAnalysis(self, dependencyGraph, reaction, classification):
        """
        adds addBond based reactions based dependencies to the dependency graph

        >>> dg = dg2 = {}
        >>> dummy = SCTSolver(None)
        >>> dummy.bindingReactionsAnalysis(dg, [['A', 'B'], ['C']], 'Binding')
        >>> dg == {'A': [], 'C': [['A', 'B']], 'B': []}
        True
        >>> dummy.bindingReactionsAnalysis(dg2, [['C'], ['A', 'B']], 'Binding')
        >>> dg2 == {'A': [], 'C': [['A', 'B']], 'B': []}
        True
        """
        totalElements = [item for sublist in reaction for item in sublist]
        for element in totalElements:
            atoAux.addToDependencyGraph(dependencyGraph, element, [])
            if classification == "Binding":
                if len(reaction[1]) == 1 and element not in reaction[0]:
                    atoAux.addToDependencyGraph(dependencyGraph, element, reaction[0])
                elif len(reaction[0]) == 1 and element not in reaction[1]:
                    atoAux.addToDependencyGraph(dependencyGraph, element, reaction[1])

    def fillSCTwithAnnotationInformation(
        self, orphanedSpecies, annotationDict, logResults=True, tentativeFlag=False
    ):
        # annotation handling
        exactMatches = defaultdict(list)
        partialMatches = defaultdict(list)
        strongIntersectionMatches = defaultdict(list)
        intersectionMatches = defaultdict(list)
        # iterate over all pairs of orphaned species
        for combinationParticle in itertools.combinations(orphanedSpecies, 2):
            # compare annotations
            if (
                combinationParticle[0] in annotationDict
                and combinationParticle[1] in annotationDict
            ):

                sortedPair = sorted(list(combinationParticle), key=len)
                # get unary keys
                unaryAnnotation1 = [
                    y
                    for x in annotationDict[combinationParticle[0]]
                    for y in annotationDict[combinationParticle[0]][x]
                    if x
                    in [
                        "BQM_IS_DESCRIBED_BY",
                        "BQB_IS_VERSION_OF",
                        "BQB_IS",
                        "BQB_ENCODES",
                    ]
                    and ("uniprot" in y or "chebi" in y)
                ]
                unaryAnnotation2 = [
                    y
                    for x in annotationDict[combinationParticle[1]]
                    for y in annotationDict[combinationParticle[1]][x]
                    if x
                    in [
                        "BQM_IS_DESCRIBED_BY",
                        "BQB_IS_VERSION_OF",
                        "BQB_IS",
                        "BQB_ENCODES",
                    ]
                    and ("uniprot" in y or "chebi" in y)
                ]

                # get compositional keys
                compositionalAnnotation1 = [
                    y
                    for x in annotationDict[combinationParticle[0]]
                    for y in annotationDict[combinationParticle[0]][x]
                    if x in ["BQB_HAS_PART", "BQB_HAS_VERSION"]
                    and ("uniprot" in y or "chebi" in y)
                ]
                compositionalAnnotation2 = [
                    y
                    for x in annotationDict[combinationParticle[1]]
                    for y in annotationDict[combinationParticle[1]][x]
                    if x in ["BQB_HAS_PART", "BQB_HAS_VERSION"]
                    and ("uniprot" in y or "chebi" in y)
                ]
                # unary keys match
                if any([x in unaryAnnotation2 for x in unaryAnnotation1]):

                    exactMatches[sortedPair[1]].append([sortedPair[0]])
                # one composes the other
                elif any([x in compositionalAnnotation1 for x in unaryAnnotation2]):
                    # if combinationParticle[0] not in partialMatches:
                    #    partialMatches[combinationParticle[0]].append([])
                    partialMatches[combinationParticle[0]].append(
                        [combinationParticle[1]]
                    )

                elif any([x in compositionalAnnotation2 for x in unaryAnnotation1]):
                    # if combinationParticle[1] not in partialMatches:
                    #    partialMatches[combinationParticle[1]].append([])
                    partialMatches[combinationParticle[1]].append(
                        [combinationParticle[0]]
                    )
                elif (
                    set(compositionalAnnotation1) == set(compositionalAnnotation2)
                    and len(
                        [
                            x in compositionalAnnotation2
                            for x in compositionalAnnotation1
                        ]
                    )
                    > 0
                ):
                    strongIntersectionMatches[sortedPair[1]].append([sortedPair[0]])
                # they intersect
                elif any(
                    [x in compositionalAnnotation2 for x in compositionalAnnotation1]
                ):

                    intersectionMatches[sortedPair[1]].append([sortedPair[0]])
                    # intersectionMatches[combinationParticle[0]].append(combinationParticle[1])
        # create unary groups

        exactMatches = self.consolidateDependencyGraph(
            dict(exactMatches), {}, {}, self.database.sbmlAnalyzer, loginformation=False
        )[0]

        if logResults:
            for x in [y for y in exactMatches if len(exactMatches[y]) > 0]:
                if not tentativeFlag:
                    logMess(
                        "INFO:ANN001",
                        "{0}:{1}: there is a direct equivalence between these sets according to annotation information. No action was taken".format(
                            x, exactMatches[x][0][0]
                        ),
                    )
                else:
                    if not (
                        x in self.database.dependencyGraph
                        and exactMatches[x][0][0] in self.database.dependencyGraph
                        and self.database.dependencyGraph[x]
                        == self.database.dependencyGraph[exactMatches[x][0][0]]
                    ):
                        logMess(
                            "WARNING:ANN101",
                            "{0}:{1} were determined to be equivalent according to annotation information. Please confirm from user information".format(
                                x, exactMatches[x][0][0]
                            ),
                        )

        # create strong intersection groups

        strongIntersectionMatches = {
            x: strongIntersectionMatches[x]
            for x in strongIntersectionMatches
            if x not in partialMatches
        }
        strongIntersectionMatches.update(exactMatches)
        strongIntersectionMatches = self.consolidateDependencyGraph(
            dict(strongIntersectionMatches),
            {},
            {},
            self.database.sbmlAnalyzer,
            loginformation=False,
        )[0]
        if logResults:
            for x in [
                y
                for y in strongIntersectionMatches
                if len(strongIntersectionMatches[y]) > 0
            ]:
                if x not in exactMatches:
                    if not tentativeFlag:
                        logMess(
                            "INFO:ANN002",
                            "{0}: can exactly match {1} according to annotation information. No action was taken".format(
                                x, strongIntersectionMatches[x]
                            ),
                        )
                    else:
                        if not (
                            x in self.database.dependencyGraph
                            and strongIntersectionMatches[x][0][0]
                            in self.database.dependencyGraph
                            and self.database.dependencyGraph[x]
                            == self.database.dependencyGraph[
                                strongIntersectionMatches[x][0][0]
                            ]
                        ):
                            logMess(
                                "WARNING:ANN101",
                                "{0}: was determined to exactly match {1} according to annotation information. Please confirm from user information".format(
                                    x, strongIntersectionMatches[x]
                                ),
                            )
        # create partial intersection groups
        """
        intersectionMatches = {x: intersectionMatches[x] for x in intersectionMatches if x not in partialMatches and x not in strongIntersectionMatches}
        intersectionMatches.update(exactMatches)


        intersectionMatches = consolidateDependencyGraph(dict(intersectionMatches), {}, {}, self.database.sbmlAnalyzer, self.database, loginformation=False)[0]
        if logResults:
            for x in intersectionMatches:
                if x not in exactMatches:
                    logMess('INFO:ANN002', '{0}: was determined to be partially match {1} according to annotation information.'.format(
                        x, intersectionMatches[x]))

        partialMatches = consolidateDependencyGraph(
            dict(partialMatches), {}, {}, self.database.sbmlAnalyzer, self.database, loginformation=False)[0]

        if logResults:
            for x in partialMatches:
                if partialMatches[x] != []:
                    logMess('INFO:ANN003', '{0}: is thought to be partially composed of reactants {1} according to annotation information. Please verify stoichiometry'.format(
                        x, partialMatches[x]))

        # validAnnotationPairs.sort()

        intersectionMatches.update(strongIntersectionMatches)
        """
        return intersectionMatches, partialMatches

    def consolidateDependencyGraph(
        self,
        dependencyGraph,
        equivalenceTranslator,
        equivalenceDictionary,
        sbmlAnalyzer,
        loginformation=True,
    ):
        """
        The second part of the Atomizer algorithm, once the lexical and stoichiometry information has been extracted
        it is time to state all elements of the system in unequivocal terms of their molecule types
        """

        equivalenceTranslator = {}

        def selectBestCandidate(
            reactant,
            candidates,
            dependencyGraph,
            sbmlAnalyzer,
            equivalenceTranslator=equivalenceTranslator,
            equivalenceDictionary=equivalenceDictionary,
        ):

            tmpCandidates = []
            modifiedElementsPerCandidate = []
            unevenElements = []
            candidateDict = {}
            for individualAnswer in candidates:
                try:
                    tmpAnswer = []
                    flag = True
                    if len(individualAnswer) == 1 and individualAnswer[0] == reactant:
                        continue
                    modifiedElements = []
                    for chemical in individualAnswer:

                        # we cannot handle tuple naming conventions for now
                        if type(chemical) == tuple:
                            flag = False
                            continue
                        # associate elements in the candidate description with their
                        # modified version
                        rootChemical = self.resolveDependencyGraph(
                            dependencyGraph, chemical
                        )
                        mod = self.resolveDependencyGraph(
                            dependencyGraph, chemical, True
                        )
                        if mod != []:
                            modifiedElements.extend(mod)
                        for element in rootChemical:
                            if len(element) == 1 and type(element[0]) == tuple:
                                continue
                            if element == chemical:
                                tmpAnswer.append(chemical)
                            elif type(element) == tuple:
                                tmpAnswer.append(element)
                            else:
                                tmpAnswer.append(element[0])
                    modifiedElementsPerCandidate.append(modifiedElements)
                    if flag:
                        tmpAnswer = sorted(tmpAnswer)
                        tmpCandidates.append(tmpAnswer)
                except atoAux.CycleError:
                    if loginformation:
                        logMess(
                            "ERROR:SCT221",
                            "{0}:{1}:Dependency cycle found when mapping molecule to candidate".format(
                                reactant, individualAnswer[0]
                            ),
                        )
                    continue
            # we cannot handle tuple naming conventions for now
            if len(tmpCandidates) == 0:
                # logMess('CRITICAL:Atomization', 'I dont know how to process these candidates and I have no \
                # way to make an educated guess. Politely refusing to translate
                # {0}={1}.'.format(reactant, candidates))
                return None, None, None
            originalTmpCandidates = deepcopy(tmpCandidates)
            # if we have more than one modified element for a single reactant
            # we can try to  choose the one that is most similar to the original
            # reactant
            # FIXME:Fails if there is a double modification
            newModifiedElements = {}
            # modifiedElementsCounter = Counter()
            modifiedElementsCounters = [Counter() for x in range(len(candidates))]
            # keep track of how many times we need to modify elements in the candidate description
            # FIXME: This only keeps track of the stuff in the fist candidates list
            for idx, modifiedElementsInCandidate in enumerate(
                modifiedElementsPerCandidate
            ):
                for element in modifiedElementsInCandidate:
                    if element[0] not in newModifiedElements or element[1] == reactant:
                        newModifiedElements[element[0]] = element[1]
                    modifiedElementsCounters[idx][element[0]] += 1

            # actually modify elements and store final version in tmpCandidates
            # if tmpCandidates[1:] == tmpCandidates[:-1] or len(tmpCandidates) ==
            # 1:

            for tmpCandidate, modifiedElementsCounter in zip(
                tmpCandidates, modifiedElementsCounters
            ):
                flag = True
                while flag:
                    flag = False
                    for idx, chemical in enumerate(tmpCandidate):
                        if modifiedElementsCounter[chemical] > 0:
                            modifiedElementsCounter[chemical] -= 1
                            tmpCandidate[idx] = newModifiedElements[chemical]
                            flag = True
                            break
            candidateDict = {tuple(x): y for x, y in zip(tmpCandidates, candidates)}
            bcan = []
            btmp = []
            borig = []
            # filter out those dependencies to the 0 element

            # if this is related to the zero element
            if len(tmpCandidates) == 1 and tmpCandidates[0] == ["0"]:
                return ["0"], None, None

            for candidate, tmpcandidate, originaltmpcandidate in zip(
                candidates, tmpCandidates, originalTmpCandidates
            ):
                if originaltmpcandidate != ["0"]:
                    bcan.append(candidate)
                    btmp.append(tmpcandidate)
                    borig.append(originaltmpcandidate)
            candidates = bcan
            tmpCandidates = btmp
            originalTmpCandidates = borig

            if len(tmpCandidates) == 0:
                return None, None, None

            # FIXME: I have no idea wtf this is doing so im commenting it out. i
            # think it's old code that is no longer ncessary
            """
            # update candidate chemical references to their modified version if required
            if len(tmpCandidates) > 1:
                # temporal solution for defaulting to the first alternative
                totalElements = [y for x in tmpCandidates for y in x]
                elementDict = {}
                for word in totalElements:
                    if word not in elementDict:
                        elementDict[word] = 0
                    elementDict[word] += 1
                newTmpCandidates = [[]]
                for element in elementDict:
                    if elementDict[element] % len(tmpCandidates) == 0:
                        newTmpCandidates[0].append(element)
                    #elif elementDict[element] % len(tmpCandidates) != 0 and re.search('(_|^){0}(_|$)'.format(element),reactant):
                    #    newTmpCandidates[0].append(element)
                    #    unevenElements.append([element])
                    else:
                        logMess('WARNING:Atomization', 'Are these actually the same? {0}={1}.'.format(reactant,candidates))
                        unevenElements.append(element)
                flag = True
                # FIXME:this should be done on newtmpCandidates instead of tmpcandidates
                while flag:
                    flag = False
                    for idx, chemical in enumerate(tmpCandidates[0]):
                        if chemical in newModifiedElements: #and newModifiedElements[chemical] in reactant:
                            tmpCandidates[0][idx] = newModifiedElements[chemical]
                            flag = True
                            break
            """
            # if all the candidates are about modification changes to a complex
            # then try to do it through lexical analysis
            if (
                all([len(candidate) == 1 for candidate in candidates])
                and candidates[0][0] != reactant
                and len(tmpCandidates[0]) > 1
            ):
                if reactant is not None:
                    pass

                # analyze based on standard modifications
                # lexCandidate, translationKeys, tmpequivalenceTranslator = sbmlAnalyzer.analyzeSpeciesModification(candidates[0][0], reactant, originalTmpCandidates[0])
                # print '++++'
                (
                    lexCandidate,
                    translationKeys,
                    tmpequivalenceTranslator,
                ) = sbmlAnalyzer.analyzeSpeciesModification2(
                    candidates[0][0], reactant, originalTmpCandidates[0]
                )
                # lexCandidate, translationKeys, tmpequivalenceTranslator = sbmlAnalyzer.analyzeSpeciesModification(candidates[0][0], reactant, tmpCandidates[0])            # FIXME: this is iffy. is it always an append modification? could be prepend
                # lexCandidate = None
                if lexCandidate is not None:
                    lexCandidate = tmpCandidates[0][
                        originalTmpCandidates[0].index(lexCandidate)
                    ]
                    if translationKeys[0] + lexCandidate in dependencyGraph:
                        lexCandidateModification = translationKeys[0] + lexCandidate
                    else:
                        lexCandidateModification = lexCandidate + translationKeys[0]

                    for element in tmpequivalenceTranslator:
                        if element not in equivalenceTranslator:
                            equivalenceTranslator[element] = []
                        equivalenceTranslator[element].append(
                            (lexCandidate, lexCandidateModification)
                        )
                    while lexCandidate in tmpCandidates[0]:
                        tmpCandidates[0].remove(lexCandidate)
                        tmpCandidates[0].append(lexCandidateModification)
                        break
                    if lexCandidateModification not in dependencyGraph:
                        logMess(
                            "WARNING:SCT711",
                            "While analyzing {0}={1} we discovered equivalence {2}={3}, please verify \
    this the correct behavior or provide an alternative for {0}".format(
                                reactant,
                                tmpCandidates[0],
                                lexCandidateModification,
                                lexCandidate,
                            ),
                        )
                    dependencyGraph[lexCandidateModification] = [[lexCandidate]]

                    return [tmpCandidates[0]], unevenElements, candidates

                else:
                    fuzzyCandidateMatch = None
                    """
                    if nothing else works and we know the result is a bimolecular
                    complex and we know which are the basic reactants then try to
                    do fuzzy string matching between the two.
                    TODO: extend this to more than 2 molecule complexes.
                    """
                    if len(tmpCandidates[0]) == 2:
                        tmpmolecules = []
                        tmpmolecules.extend(originalTmpCandidates[0])
                        tmpmolecules.extend(tmpCandidates[0])
                        # FIXME: Fuzzy artificial reaction is using old methods. Try to fix this
                        # or maybe not, no one was using it and when it was used it was wrong
                        # fuzzyCandidateMatch = sbmlAnalyzer.fuzzyArtificialReaction(originalTmpCandidates[0],[reactant],tmpmolecules)
                        fuzzyCandidateMatch = None
                    if fuzzyCandidateMatch is not None:
                        # logMess('INFO:Atomization', 'Used fuzzy string matching from {0} to {1}'.format(reactant, fuzzyCandidateMatch))
                        return [fuzzyCandidateMatch], unevenElements, candidates
                    else:
                        # map based on greedy matching
                        greedyMatch = sbmlAnalyzer.greedyModificationMatching(
                            reactant, dependencyGraph.keys()
                        )
                        if greedyMatch not in [-1, -2]:
                            return (
                                selectBestCandidate(
                                    reactant,
                                    [greedyMatch],
                                    dependencyGraph,
                                    sbmlAnalyzer,
                                )[0],
                                unevenElements,
                                candidates,
                            )

                        # last ditch attempt using straighforward lexical analysis
                        (
                            tmpDependency,
                            tmpEquivalence,
                        ) = sbmlAnalyzer.findClosestModification(
                            [reactant],
                            dependencyGraph.keys(),
                            self.database.annotationDict,
                            self.database.dependencyGraph,
                        )
                        if (
                            reactant in tmpDependency
                            and tmpDependency[reactant] in tmpCandidates[0]
                        ):
                            for element in tmpDependency:
                                if element not in dependencyGraph:
                                    dependencyGraph[element] = tmpDependency[element]
                            for element in tmpEquivalence:
                                if element not in equivalenceDictionary:
                                    equivalenceDictionary[element] = []
                                for equivalence in tmpEquivalence[element]:
                                    if (
                                        equivalence[0]
                                        not in equivalenceDictionary[element]
                                    ):
                                        equivalenceDictionary[element].append(
                                            equivalence[0]
                                        )
                            if len(tmpDependency.keys()) > 0:
                                return (
                                    tmpDependency[reactant],
                                    unevenElements,
                                    candidates,
                                )
                        # XXX: be careful of this change. This basically forces changes to happen
                        # the ive no idea whats going on branch
                        # modificationCandidates = {}
                        # if modificationCandidates == {}:

                        activeCandidates = []
                        for individualCandidate in tmpCandidates:
                            for tmpCandidate in individualCandidate:
                                activeQuery = None
                                uniprotkey = atoAux.getURIFromSBML(
                                    tmpCandidate, self.database.parser, ["uniprot"]
                                )
                                if len(uniprotkey) > 0:
                                    uniprotkey = uniprotkey[0].split("/")[-1]
                                    activeQuery = pwcm.queryActiveSite(uniprotkey, None)
                                if activeQuery and len(activeQuery) > 0:
                                    activeCandidates.append(tmpCandidate)
                                    # enter modification information to self.database
                                    # logMess('INFO:SCT051', '{0}:Determined that {0} has an active site for modication'.format(reactant, tmpCandidate))
                                    # return [individualCandidate], unevenElements, candidates
                                # we want relevant biological names, its useless if they are too short
                                elif len(tmpCandidate) >= 3:
                                    # else:
                                    individualMajorCandidates = [
                                        y for x in candidates for y in x
                                    ]
                                    activeQuery = pwcm.queryActiveSite(
                                        tmpCandidate, None
                                    )
                                    if activeQuery and len(activeQuery) > 0:
                                        otherMatches = [
                                            x
                                            for x in tmpCandidates[0]
                                            if x in activeQuery
                                        ]
                                        if any(
                                            [
                                                x
                                                for x in otherMatches
                                                if len(x) > len(tmpCandidate)
                                            ]
                                        ):
                                            continue
                                        activeCandidates.append(tmpCandidate)
                                    # enter modification information to self.database
                                    # logMess('INFO:SCT051', '{0}:Determined that {1} has an active site for modication'.format(reactant, tmpCandidate))
                                    # return [individualCandidate], unevenElements, candidates
                        if len(activeCandidates) > 0:
                            if len(activeCandidates) == 1:
                                logMess(
                                    "INFO:SCT051",
                                    "{0}:Determined through uniprot active site query that {1} has an active site for modication".format(
                                        reactant, activeCandidates[0]
                                    ),
                                )
                            if len(activeCandidates) > 1:
                                logMess(
                                    "WARNING:SCT151",
                                    "{0}:Determined through uniprot active site query that {1} have active site for modication. Defaulting to {2}".format(
                                        reactant, activeCandidates, activeCandidates[0]
                                    ),
                                )

                            for tmpCandidate, candidate in zip(
                                tmpCandidates, candidates
                            ):
                                fuzzyList = sbmlAnalyzer.processAdHocNamingConventions(
                                    reactant,
                                    candidate[0],
                                    {},
                                    False,
                                    dependencyGraph.keys(),
                                )
                                if len(fuzzyList) > 0 and fuzzyList[0][1]:
                                    if sbmlAnalyzer.testAgainstExistingConventions(
                                        fuzzyList[0][1],
                                        sbmlAnalyzer.namingConventions[
                                            "modificationList"
                                        ],
                                    ):
                                        self.database.eequivalenceTranslator2[
                                            fuzzyList[0][1]
                                        ].append(
                                            (
                                                activeCandidates[0],
                                                "{0}{1}".format(
                                                    activeCandidates, fuzzyList[0][1]
                                                ),
                                            )
                                        )
                                    else:
                                        self.database.eequivalenceTranslator2[
                                            fuzzyList[0][1]
                                        ] = [
                                            (
                                                activeCandidates[0],
                                                "{0}{1}".format(
                                                    activeCandidates[0], fuzzyList[0][1]
                                                ),
                                            )
                                        ]

                                    if (
                                        "{0}{1}".format(
                                            activeCandidates[0], fuzzyList[0][1]
                                        )
                                        not in dependencyGraph
                                    ):
                                        dependencyGraph[
                                            "{0}{1}".format(
                                                activeCandidates[0], fuzzyList[0][1]
                                            )
                                        ] = [[activeCandidates[0]]]

                                    for idx, element in enumerate(tmpCandidate):
                                        if element == activeCandidates[0]:
                                            tmpCandidates[0][idx] = "{0}{1}".format(
                                                activeCandidates[0], fuzzyList[0][1]
                                            )
                                            break
                                    return (
                                        [tmpCandidates[0]],
                                        unevenElements,
                                        candidates,
                                    )

                        if len(tmpCandidates) != 1:
                            if not self.database.softConstraints:
                                if loginformation:
                                    logMess(
                                        "ERROR:SCT213",
                                        "{0}:Atomizer needs user information to determine which element is being modified among components {1}={2}.".format(
                                            reactant, candidates, tmpCandidates
                                        ),
                                    )
                                # print self.database.userLabelDictionary
                                return None, None, None
                        else:

                            if not self.database.softConstraints:
                                if loginformation:
                                    modification = (
                                        sbmlAnalyzer.findMatchingModification(
                                            reactant, candidates[0][0]
                                        )
                                    )
                                    modification = (
                                        modification[0] if modification else "mod"
                                    )
                                    logMess(
                                        "ERROR:SCT212",
                                        "{1}:{0}:Atomizer needs user information to determine which element is being modified among component species:{2}:{3}".format(
                                            reactant,
                                            candidates[0],
                                            tmpCandidates[0],
                                            modification,
                                        ),
                                    )

                                return None, None, None

                        # return [tmpCandidates[0]], unevenElements

            elif len(tmpCandidates) > 1:
                # all candidates are equal/consistent
                if all(sorted(x) == sorted(tmpCandidates[0]) for x in tmpCandidates):
                    tmpCandidates = [tmpCandidates[0]]
                elif (
                    reactant in self.database.alternativeDependencyGraph
                    and loginformation
                ):
                    # candidates contradict each other but we have naming convention information in alternativeDependencyGraph
                    if not all(
                        sorted(x) == sorted(originalTmpCandidates[0])
                        for x in originalTmpCandidates
                    ):
                        if loginformation:
                            logMess(
                                "INFO:SCT001",
                                "{0}:Using lexical analysis since stoichiometry gives non-consistent information naming({1})!=stoichiometry({2})".format(
                                    reactant,
                                    self.database.alternativeDependencyGraph[reactant][
                                        0
                                    ],
                                    tmpCandidates,
                                ),
                            )

                    # else:
                    #    print self.database.alternativeDependencyGraph[reactant],tmpCandidates,reactant
                    #    logMess('INFO:Atomization', 'Using lexical analysis for species {0} =  {1} since stoichiometry gave conflicting information {2}'.format(reactant,
                    # self.database.alternativeDependencyGraph[reactant][0],
                    # tmpCandidates))

                    # fallback to naming conventions
                    candidate = self.database.alternativeDependencyGraph[reactant]
                    # resolve naming convention candidate to its basic components
                    # (molecule types)
                    namingTmpCandidates = selectBestCandidate(
                        reactant, [candidate[0]], dependencyGraph, sbmlAnalyzer
                    )[0]
                    if not namingTmpCandidates:
                        logMess(
                            "ERROR:SCT211",
                            "{0}:{1}:{2}:Cannot converge to solution, conflicting definitions".format(
                                reactant, tmpCandidates, originalTmpCandidates
                            ),
                        )
                        return None, None, None
                    if not any(
                        [
                            sorted(subcandidate) == sorted(namingTmpCandidates[0])
                            for subcandidate in tmpCandidates
                        ]
                    ):
                        if loginformation:
                            logMess(
                                "WARNING:SCT112",
                                "{0}:Stoichiometry analysis:{1}:results in non self-consistent definitions and conflicts with lexical analysis:{2}:Selecting lexical analysis".format(
                                    reactant, tmpCandidates, namingTmpCandidates
                                ),
                            )
                        atoAux.addAssumptions(
                            "lexicalVsstoch",
                            (
                                reactant,
                                ("lexical", str(namingTmpCandidates)),
                                ("stoch", str(tmpCandidates)),
                                ("original", str(originalTmpCandidates)),
                            ),
                            self.database.assumptions,
                        )

                    tmpCandidates = namingTmpCandidates
                    if loginformation:
                        self.database.alternativeDependencyGraph[
                            reactant
                        ] = tmpCandidates
                elif all(
                    sorted(x) == sorted(originalTmpCandidates[0])
                    for x in originalTmpCandidates
                ):
                    # the basic elements are the same but we are having trouble matching modifciations together
                    sortedCandidates = sorted(
                        [
                            ([y for y in x if y in reactant], i)
                            for i, x in enumerate(tmpCandidates)
                        ],
                        key=lambda z: [len(z[0]), sum([len(w) for w in z[0]])],
                        reverse=True,
                    )
                    if loginformation:
                        logMess(
                            "WARNING:SCT113",
                            "{0}:candidates:{1}:agree on the basic components but naming conventions cannot determine  specific modifications. Selecting:{2}:based on longest partial match".format(
                                reactant,
                                tmpCandidates,
                                tmpCandidates[sortedCandidates[0][1]],
                            ),
                        )
                    replacementCandidate = [tmpCandidates[sortedCandidates[0][1]]]
                    atoAux.addAssumptions(
                        "lexicalVsstoch",
                        (
                            reactant,
                            ("current", str(replacementCandidate)),
                            (
                                "alternatives",
                                str(
                                    [
                                        x
                                        for x in tmpCandidates
                                        if x != replacementCandidate[0]
                                    ]
                                ),
                            ),
                            ("original", str(originalTmpCandidates)),
                        ),
                        self.database.assumptions,
                    )
                    tmpCandidates = replacementCandidate
                else:
                    tmpCandidates2 = [
                        x
                        for x in tmpCandidates
                        if all(y not in x for y in self.database.constructedSpecies)
                    ]
                    # if we had constructed species disregard those since they are introducing noise
                    if len(tmpCandidates2) > 0 and len(tmpCandidates) != len(
                        tmpCandidates2
                    ):
                        return selectBestCandidate(
                            reactant, tmpCandidates2, dependencyGraph, sbmlAnalyzer
                        )
                    elif len(tmpCandidates2) == 0:
                        # the differences is between species that we created so its the LAE fault. Just choose one.
                        tmpCandidates.sort(key=len)
                        tmpCandidates = [tmpCandidates[0]]
                    else:
                        if loginformation:
                            logMess(
                                "ERROR:SCT211",
                                "{0}:{1}:{2}:Cannot converge to solution, conflicting definitions".format(
                                    reactant, tmpCandidates, originalTmpCandidates
                                ),
                            )
                        return None, None, None
            elif (
                reactant in self.database.alternativeDependencyGraph and loginformation
            ):
                # there is one stoichionetry candidate but the naming convention
                # and the stoichionetry dotn agree
                if (
                    tmpCandidates[0]
                    != self.database.alternativeDependencyGraph[reactant][0]
                ):
                    # make sure the naming convention is resolved to basic
                    # omponents
                    candidate = self.database.alternativeDependencyGraph[reactant]
                    # this is to avoid recursion
                    if loginformation:
                        del self.database.alternativeDependencyGraph[reactant]
                    namingtmpCandidates = selectBestCandidate(
                        reactant, [candidate[0]], dependencyGraph, sbmlAnalyzer
                    )[0]

                    # if they still disagree print error and use stoichiometry
                    if (
                        namingtmpCandidates
                        and tmpCandidates[0] != namingtmpCandidates[0]
                    ):

                        if loginformation:
                            if (
                                namingtmpCandidates[0][0]
                                in self.database.constructedSpecies
                            ):
                                namingTmpCandidates = tmpCandidates

                            else:
                                self.database.alternativeDependencyGraph[
                                    reactant
                                ] = namingtmpCandidates
                                logMess(
                                    "WARNING:SCT111",
                                    "{0}:stoichiometry analysis:{1}:conflicts with and naming conventions:{2}:Selecting lexical analysis".format(
                                        reactant,
                                        tmpCandidates,
                                        self.database.alternativeDependencyGraph[
                                            reactant
                                        ],
                                    ),
                                )
                        tmpCandidates = namingtmpCandidates
                        atoAux.addAssumptions(
                            "lexicalVsstoch",
                            (
                                reactant,
                                ("stoch", str(tmpCandidates)),
                                ("lexical", str(namingtmpCandidates)),
                                ("original", str(originalTmpCandidates)),
                            ),
                            self.database.assumptions,
                        )
                        for element in tmpCandidates[0]:
                            if element not in prunnedDependencyGraph:
                                # elemental species that were not used anywhere
                                # else but for those entries discovered through
                                # naming conventions
                                prunnedDependencyGraph[element] = []
                    elif not namingtmpCandidates:
                        if loginformation:
                            logMess(
                                "WARNING:SCT121",
                                "{0}:could not resolve naming({1}) into a viable compositional candidate. choosing stoichiometry({2})".format(
                                    reactant, candidate, tmpCandidates[0]
                                ),
                            )
            originalCandidateName = (
                candidateDict[tuple(tmpCandidates[0])]
                if tuple(tmpCandidates[0]) in candidateDict
                else None
            )
            return [tmpCandidates[0]], unevenElements, originalCandidateName

        prunnedDependencyGraph = deepcopy(dependencyGraph)

        tempMergedDependencyGraph = deepcopy(prunnedDependencyGraph)
        for element in self.database.alternativeDependencyGraph:
            if element in tempMergedDependencyGraph:
                tempMergedDependencyGraph[element].extend(
                    self.database.alternativeDependencyGraph[element]
                )
        weights = self.weightDependencyGraph(tempMergedDependencyGraph)

        # raise Exception

        unevenElementDict = {}
        for element in weights:
            candidates = [x for x in prunnedDependencyGraph[element[0]]]
            if len(candidates) == 1 and type(candidates[0][0]) == tuple:
                prunnedDependencyGraph[element[0]] = []
            if len(candidates) >= 1:
                candidates, uneven, originalCandidate = selectBestCandidate(
                    element[0], candidates, prunnedDependencyGraph, sbmlAnalyzer
                )
                # except CycleError:
                #    candidates = None
                #    uneven = []
                if uneven != []:
                    unevenElementDict[element[0]] = uneven
            if candidates is None:
                prunnedDependencyGraph[element[0]] = []
            else:
                prunnedDependencyGraph[element[0]] = [sorted(x) for x in candidates]

        weights = self.weightDependencyGraph(prunnedDependencyGraph)
        return prunnedDependencyGraph, weights, unevenElementDict, equivalenceTranslator

    @memoize
    def measureGraph(self, element, path):
        """
        Calculates the weight of individual paths as the sum of the weights of the individual candidates plus the number of
        candidates. The weight of an individual candidate is equal to the sum of strings contained in that candidate different
        from the original reactant
        >>> dummy = SCTSolver(None)
        >>> dummy.measureGraph('Trash',['0'])
        1
        >>> dummy.measureGraph('EGF',[['EGF']])
        2
        >>> dummy.measureGraph('EGFR_P',[['EGFR']])
        3
        >>> dummy.measureGraph('EGF_EGFR', [['EGF', 'EGFR']])
        4
        >>> dummy.measureGraph('A_B_C',[['A', 'B_C'], ['A_B', 'C']])
        7
        """
        counter = 1
        for x in path:
            if type(x) == list or type(x) == tuple:
                counter += self.measureGraph(element, x)
            elif x != "0" and x != element:
                counter += 1
        return counter

    # ASS: From my testing the iterative version is not only identical
    # but also significantly faster for most models since measure
    # graph doesn't get the same inputs, memoization doesn't pay off.
    def measureGraph2(self, element, path):
        """
        Identical to previous function but iterative instead of
        recursive
        """
        counter = 1
        if len(path) == 1:
            if type(path[0]) == list or type(path[0]) == tuple:
                counter += 1
                # check inside
                for x in path[0]:
                    if x != "0" and x != element:
                        counter += 1
            else:
                if path[0] != "0" and path[0] != element:
                    counter += 1
        else:
            # it's a longer thing
            counter += len(path)
            # flatten and check
            flat = [i for sb in path for i in sb if i]
            for x in flat:
                if x != "0" and x != element:
                    counter += 1
        return counter

    def weightDependencyGraph(self, dependencyGraph):
        """
        Given a dependency Graph it will return a list indicating the weights of its elements
        a path is calculated 
        >>> dummy = SCTSolver(None)
        >>> dummy.weightDependencyGraph({'EGF_EGFR_2':[['EGF_EGFR','EGF_EGFR']],'EGF_EGFR':[['EGF','EGFR']],'EGFR':[],'EGF':[],\
        'EGFR_P':[['EGFR']],'EGF_EGFR_2_P':[['EGF_EGFR_2']]})
        [['EGF', 2], ['EGFR', 2], ['EGFR_P', 4], ['EGF_EGFR', 5], ['EGF_EGFR_2', 9], ['EGF_EGFR_2_P', 10]]
        >>> dependencyGraph2 = {'A':[],'B':[],'C':[],'A_B':[['A','B']],'B_C':[['B','C']],'A_B_C':[['A_B','C'],['B_C','A']]}
        >>> dummy.weightDependencyGraph(dependencyGraph2)
        [['A', 2], ['C', 2], ['B', 2], ['B_C', 5], ['A_B', 5], ['A_B_C', 13]]
        """
        weights = []
        for element in dependencyGraph:
            path = self.resolveDependencyGraph(dependencyGraph, element)
            try:
                path2 = self.resolveDependencyGraph(dependencyGraph, element, True)
            except atoAux.CycleError:
                path2 = []
            # ASS: Swapping to iterative version of the function
            # weight = self.measureGraph(element, path) + len(path2)
            weight = self.measureGraph2(element, path) + len(path2)
            weights.append([element, weight])

        weights = sorted(weights, key=lambda rule: (rule[1], len(rule[0])))
        return weights

    # ASS: New method to make hashes from graphs. Some key points
    # 1) sorting is done to ensure same graph gives the same key, consistently
    # 2) python internal hashing function is used for the hashing
    # 3) should be very collision proof
    def make_key_from_graph(self, graph):
        hashable_tuples = []
        # If graph is empty just return the empty tuple result
        if len(graph) == 0:
            return marshal.dumps(hashable_tuples)
        # So we don't modify original graph
        tmpGraph = deepcopy(graph)
        # This turns the graph into a traditional graph implementation
        # where there are no edges that go to nodes that do not exist in the
        # graph, I'm making sure every node exists in the graph itself
        all_elems = set(
            [item[0] for sublist in tmpGraph.values() for item in sublist if item]
        )
        for elem in all_elems:
            try:
                a = tmpGraph[elem]
            except KeyError:
                tmpGraph[elem] = []
        # Now we should have a traditional graph implementation
        # I also want to unroll every element to turn this into a hashable
        # tuple of tuples type deal
        for key in sorted(tmpGraph):
            tmpGraph[key] = functools.reduce(
                lambda x, y: x + y, sorted(tmpGraph[key]), []
            )
        # Now we can turn this into a proper hashable object
        for key in sorted(tmpGraph):
            hashable_tuples.append((key, tuple(tmpGraph[key])))
        # Turn the list into tuples to it's hashable
        hashable_tuples = tuple(hashable_tuples)
        # return hash
        return hashable_tuples.__hash__()

    def resolveDependencyGraph(
        self, dependencyGraph, reactant, withModifications=False
    ):
        """
        Given a full species composition table and a reactant it will return an unrolled list of the molecule types
        (elements with no dependencies that define this reactant). The classification to the original candidates is lost
        since elements are fully unrolled. For getting dependencies keeping candidate consistency use consolidateDependencyGraph
        instead
        
        Args:
            withModifications (bool): returns a list of the 1:1 transformation relationships found in the path to this graph

        >>> dummy = SCTSolver(None)
        >>> dependencyGraph = {'EGF_EGFR_2':[['EGF_EGFR','EGF_EGFR']],'EGF_EGFR':[['EGF','EGFR']],'EGFR':[],'EGF':[],\
        'EGFR_P':[['EGFR']],'EGF_EGFR_2_P':[['EGF_EGFR_2']]}
        >>> dependencyGraph2 = {'A':[],'B':[],'C':[],'A_B':[['A','B']],'B_C':[['B','C']],'A_B_C':[['A_B','C'],['B_C','A']]}
        >>> dummy.resolveDependencyGraph(dependencyGraph, 'EGFR')
        [['EGFR']]
        >>> dummy.resolveDependencyGraph(dependencyGraph, 'EGF_EGFR')
        [['EGF'], ['EGFR']]
        >>> sorted(dummy.resolveDependencyGraph(dependencyGraph, 'EGF_EGFR_2_P'))
        [['EGF'], ['EGF'], ['EGFR'], ['EGFR']]
        
        >>> sorted(dummy.resolveDependencyGraph(dependencyGraph, 'EGF_EGFR_2_P', withModifications=True))
        [('EGF_EGFR_2', 'EGF_EGFR_2_P')]
        >>> sorted(dummy.resolveDependencyGraph(dependencyGraph2,'A_B_C'))
        [['A'], ['A'], ['B'], ['B'], ['C'], ['C']]
        """
        gkey = self.make_key_from_graph(dependencyGraph)
        try:
            self.dg = self.graph_map[gkey]
        except KeyError:
            self.graph_map[gkey] = dependencyGraph
            self.dg = dependencyGraph

        if self.memoizedResolver:
            topCandidate = self.resolveDependencyGraphHelper(
                gkey, reactant, [], withModifications
            )
        else:
            topCandidate = self.unMemoizedResolveDependencyGraphHelper(
                self.dg, reactant, [], withModifications
            )
        return topCandidate

    @memoizeMapped
    def resolveDependencyGraphHelper(
        self, gkey, reactant, memory, withModifications=False
    ):
        """
        Helper function for resolveDependencyGraph that adds a memory field to resolveDependencyGraphHelper to avoid 
        cyclical definitions problems 
        >>> dummy = SCTSolver(None)
        >>> dependencyGraph = {'EGF_EGFR_2':[['EGF_EGFR','EGF_EGFR']],'EGF_EGFR':[['EGF','EGFR']],'EGFR':[],'EGF':[],\
        'EGFR_P':[['EGFR']],'EGF_EGFR_2_P':[['EGF_EGFR_2']]}
        >>> dependencyGraph2 = {'A':[],'B':[],'C':[],'A_B':[['A','B']],'B_C':[['B','C']],'A_B_C':[['A_B','C'],['B_C','A']]}
        >>> sorted(dummy.resolveDependencyGraphHelper(dependencyGraph, 'EGF_EGFR_2_P',[]))
        [['EGF'], ['EGF'], ['EGFR'], ['EGFR']]
       
        >>> sorted(dummy.resolveDependencyGraphHelper(dependencyGraph, 'EGF_EGFR_2_P', [], withModifications=True))
        [('EGF_EGFR_2', 'EGF_EGFR_2_P')]

        >>> sorted(dummy.resolveDependencyGraphHelper(dependencyGraph2, 'A_B_C', []))
        [['A'], ['A'], ['B'], ['B'], ['C'], ['C']]

        >>> dependencyGraph3 = {'C1': [['C2']],'C2':[['C3']],'C3':[['C1']]}
        >>> resolveDependencyGraphHelper(dummy.dependencyGraph3, 'C3', [], withModifications=True)
        Traceback (innermost last):
          File "<stdin>", line 1, in ?
        CycleError
        """

        result = []
        # if type(reactant) == tuple:
        #    return []
        if (
            reactant not in self.dg
            or self.dg[reactant] == []
            or self.dg[reactant] == [[reactant]]
        ):
            if not withModifications:
                result.append([reactant])
        else:
            for option in self.dg[reactant]:
                tmp = []
                for element in option:
                    if element in memory and not withModifications:
                        result.append([element])
                        continue
                    elif element in memory:
                        # logMess(
                        #    'ERROR:SCT201', 'dependency cycle detected on {0}'.format(element))
                        raise atoAux.CycleError(memory)
                    baseElement = self.resolveDependencyGraphHelper(
                        gkey, element, memory + [element], withModifications
                    )
                    if baseElement is not None:
                        tmp.extend(baseElement)
                # if not withModifications:
                result.extend(tmp)
                if len(option) == 1 and withModifications and option[0] != reactant:
                    result.append((option[0], reactant))
        return result

    def unMemoizedResolveDependencyGraphHelper(
        self, dependencyGraph, reactant, memory, withModifications=False
    ):
        """
        Helper function for resolveDependencyGraph that adds a memory field to resolveDependencyGraphHelper to avoid 
        cyclical definitions problems 
        >>> dummy = SCTSolver(None)
        >>> dependencyGraph = {'EGF_EGFR_2':[['EGF_EGFR','EGF_EGFR']],'EGF_EGFR':[['EGF','EGFR']],'EGFR':[],'EGF':[],\
        'EGFR_P':[['EGFR']],'EGF_EGFR_2_P':[['EGF_EGFR_2']]}
        >>> dependencyGraph2 = {'A':[],'B':[],'C':[],'A_B':[['A','B']],'B_C':[['B','C']],'A_B_C':[['A_B','C'],['B_C','A']]}
        >>> sorted(dummy.resolveDependencyGraphHelper(dependencyGraph, 'EGF_EGFR_2_P',[]))
        [['EGF'], ['EGF'], ['EGFR'], ['EGFR']]
       
        >>> sorted(dummy.resolveDependencyGraphHelper(dependencyGraph, 'EGF_EGFR_2_P', [], withModifications=True))
        [('EGF_EGFR_2', 'EGF_EGFR_2_P')]

        >>> sorted(dummy.resolveDependencyGraphHelper(dependencyGraph2, 'A_B_C', []))
        [['A'], ['A'], ['B'], ['B'], ['C'], ['C']]

        >>> dependencyGraph3 = {'C1': [['C2']],'C2':[['C3']],'C3':[['C1']]}
        >>> resolveDependencyGraphHelper(dummy.dependencyGraph3, 'C3', [], withModifications=True)
        Traceback (innermost last):
          File "<stdin>", line 1, in ?
        CycleError
        """

        result = []
        # if type(reactant) == tuple:
        #    return []
        if (
            reactant not in dependencyGraph
            or dependencyGraph[reactant] == []
            or dependencyGraph[reactant] == [[reactant]]
        ):
            if not withModifications:
                result.append([reactant])
        else:
            for option in dependencyGraph[reactant]:
                tmp = []
                for element in option:
                    if element in memory and not withModifications:
                        result.append([element])
                        continue
                    elif element in memory:
                        # logMess(
                        #    'ERROR:SCT201', 'dependency cycle detected on {0}'.format(element))
                        raise atoAux.CycleError(memory)
                    baseElement = self.unMemoizedResolveDependencyGraphHelper(
                        dependencyGraph, element, memory + [element], withModifications
                    )
                    if baseElement is not None:
                        tmp.extend(baseElement)
                # if not withModifications:
                result.extend(tmp)
                if len(option) == 1 and withModifications and option[0] != reactant:
                    result.append((option[0], reactant))
        return result



# 2013.10.10 11:42:06 EDT
# Embedded file name: ../utils/extractAtomic.py
"""
Created on Wed Aug  7 20:54:17 2013

@author: proto
"""
from collections import Counter


def extractMolecules(action, site1, site2, chemicalArray, differentiateDimers=False):
    """
    this method goes through the chemicals in a given array 'chemicalArray'
    and extracts its atomic patterns into two arrays:
        those elements that are contained in [site1,site2] will be put in the
        reactionCenter set. The rest will    be placed in the context set.
        The entire system will be put into the atomicPatterns dictionary
    Keyword arguments:
    site1,site2 -- where the action takes place
    chemicalArray -- the list of species we will be extracting atomic patters
    from
    """
    atomicPatterns = {}
    reactionCenter = set()
    context = set()
    reactionCenterC = Counter()
    contextC = Counter()
    for reactant in chemicalArray:
        ta, tr, tc = reactant.extractAtomicPatterns(
            action, site1, site2, differentiateDimers
        )
        atomicPatterns.update(ta)
        for element in tr:
            reactionCenter.add(element)

        for element in tc:
            context.add(element)

        reactionCenterC.update(tr)
        contextC.update(tc)

    return (atomicPatterns, reactionCenterC, contextC)


def solveWildcards(atomicArray):
    """
    When you have a wildcard '+' as a bond configuration, this method allows you
    to go through the list of atomic elements and find which patterns the '+'
    can potentially resolve to
    """
    standinArray = {}
    for wildcard in [x for x in atomicArray if "+" in x]:
        for atomic in [
            x for x in atomicArray if "+" not in x and len(atomicArray[x].molecules) > 1
        ]:
            if atomicArray[wildcard].molecules[0].name in [
                x.name for x in atomicArray[atomic].molecules
            ]:
                if wildcard not in standinArray:
                    standinArray[wildcard] = []
                standinArray[wildcard].append(atomicArray[atomic])

    atomicArray.update(standinArray)


def getMapping(mapp, site):
    for mapping in mapp:
        if site in mapping:
            return [x for x in mapping if x != site][0]


def extractTransformations(rules, differentiateDimers=False):
    """
    goes through the list of rules and extracts its reactioncenter,context and product
    atomic patterns per transformation action
    """
    atomicArray = {}
    transformationCenter = []
    transformationContext = []
    productElements = []
    actionName = []
    index = 0
    label = []
    for rule, _, reationRate, reactionSymbol in rules:
        index += 1
        for action in rule.actions:
            atomic, reactionCenter, context = extractMolecules(
                action.action,
                action.site1,
                action.site2,
                rule.reactants,
                differentiateDimers,
            )
            transformationCenter.append(reactionCenter)
            transformationContext.append(context)
            atomicArray.update(atomic)
            productSites = [
                getMapping(rule.mapping, action.site1),
                getMapping(rule.mapping, action.site2),
            ]
            atomic, rc, _ = extractMolecules(
                action.action,
                productSites[0],
                productSites[1],
                rule.products,
                differentiateDimers,
            )
            productElements.append(rc)
            atomicArray.update(atomic)
            actionName.append("%i-%s" % (index, action.action))
            r = "+".join([str(x) for x in rule.reactants])
            p = "+".join([str(x) for x in rule.products])
            label.append("->".join([r, p, "%i-%s" % (index, action.action)]))

    solveWildcards(atomicArray)
    return (
        atomicArray,
        transformationCenter,
        transformationContext,
        productElements,
        actionName,
        label,
    )


if __name__ == "__main__":
    import readBNGXML

    _, rules, _ = readBNGXML.parseXML("output48.xml")
    for idx, rule in enumerate(rules):
        (
            tatomicArray,
            ttransformationCenter,
            ttransformationContext,
            tproductElements,
            tactionNames,
            tlabelArray,
        ) = extractTransformations([rule], True)
        if rule[0].label == "_reverse_v7":
            print(str(rule[0]))
            print(ttransformationContext)
            print(tproductElements)
            raise Exception



import bionetgen.atomizer.libsbml2bngl as ls2b
from bionetgen.core.defaults import BNGDefaults
import yaml, os

from bionetgen.core.utils.logging import BNGLogger, log_level


d = BNGDefaults()


class AtomizeTool:
    def __init__(
        self, input_file=None, options_dict=None, parser_namespace=None, app=None
    ):
        self.app = app
        self.logger = BNGLogger(app=self.app)
        self.logger.debug(
            "Setting up AtomizeTool object", loc=f"{__file__} : AtomizeTool.__init__()"
        )
        # we generate our defaults first and override it with
        # the dictionary first and then the namespace
        config = {
            "input": None,  # we need this, check at the end and fail if we don't have it
            "annotation": False,
            "output": None,
            "convention_file": None,
            "naming_conventions": None,
            "user_structures": None,
            "molecule_id": False,
            "convert_units": False,  # currently not supported
            "atomize": False,  # default is flat translation
            "pathwaycommons": True,  # requires connection so default is false
            "bionetgen_analysis": os.path.join(
                d.bng_path, "BNG2.pl"
            ),  # TODO: get it from app config
            "isomorphism_check": False,  # wtf do we do here?
            "ignore": False,  # wtf do we do here?
            "memoized_resolver": False,
            "keep_local_parameters": False,
            "quiet_mode": False,
            "obs_map_file": None,
            "log_level": "DEBUG",  # options are "CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"
        }
        # input file
        if input_file is not None:
            config["input"] = input_file
        # dictionary override
        if options_dict is not None:
            for key in config:
                if key in options_dict:
                    config[key] = options_dict[key]
        # namespace override
        if parser_namespace is not None:
            for key in config:
                if hasattr(parser_namespace, key):
                    config[key] = getattr(parser_namespace, key)
        # special handling of log level
        if log_level is not None:
            config["log_level"] = log_level
        elif self.app is not None:
            # we called this from the CLI
            if self.app.pargs.debug:
                config["log_level"] = "DEBUG"
            elif self.app.pargs.log_level is not None:
                config["log_level"] = self.app.pargs.log_level
            else:
                # we called the CLI but didn't give any log_level info
                config["log_level"] = "INFO"
        else:
            # we called from library but no log_level info exists
            config["log_level"] = "INFO"

        # check config options
        self.config = self.checkConfig(config)

    def checkConfig(self, config):
        self.logger.debug(
            "Validating config options", loc=f"{__file__} : AtomizeTool.checkConfig()"
        )
        options = {}
        options["inputFile"] = config["input"]  # TODO: ensure this is not None
        conv, useID, naming = ls2b.selectReactionDefinitions(options["inputFile"])
        options["outputFile"] = (
            config["output"]
            if config["output"] is not None
            else options["inputFile"] + ".bngl"
        )
        options["conventionFile"] = (
            config["convention_file"] if config["convention_file"] is not None else conv
        )
        options["userStructure"] = config["user_structures"]
        options["namingConventions"] = (
            config["naming_conventions"]
            if config["naming_conventions"] is not None
            else naming
        )
        options["useId"] = config["molecule_id"]
        options["annotation"] = config["annotation"]
        options["atomize"] = config["atomize"]
        options["pathwaycommons"] = config["pathwaycommons"]
        options["bionetgenAnalysis"] = config["bionetgen_analysis"]
        options["isomorphismCheck"] = config["isomorphism_check"]
        options["ignore"] = config["ignore"]
        options["noConversion"] = not config["convert_units"]
        options["memoizedResolver"] = config["memoized_resolver"]
        options["replaceLocParams"] = not config["keep_local_parameters"]
        options["quietMode"] = config["quiet_mode"]
        options["obs_map_file"] = config["obs_map_file"]
        assert config["log_level"] in [
            "CRITICAL",
            "ERROR",
            "WARNING",
            "INFO",
            "DEBUG",
        ], "Logging level {} is not an allowed level".format(config["log_level"])
        options["logLevel"] = config["log_level"]
        return options

    def run(self):
        # TODO: Make atomizer also use cement app logging
        # this involves changing a lot of code in atomizer!
        self.logger.debug("Analyzing SBML file", loc=f"{__file__} : AtomizeTool.run()")
        self.returnArray = ls2b.analyzeFile(
            self.config["inputFile"],
            self.config["conventionFile"],
            self.config["useId"],
            self.config["namingConventions"],
            self.config["outputFile"],
            speciesEquivalence=self.config["userStructure"],
            atomize=self.config["atomize"],
            bioGrid=False,
            pathwaycommons=self.config["pathwaycommons"],
            ignore=self.config["ignore"],
            noConversion=self.config["noConversion"],
            memoizedResolver=self.config["memoizedResolver"],
            replaceLocParams=self.config["replaceLocParams"],
            quietMode=self.config["quietMode"],
            logLevel=self.config["logLevel"],
            obs_map_file=self.config["obs_map_file"],
            app=self.app,
        )
        self.logger.debug("Post-analysis", loc=f"{__file__} : AtomizeTool.run()")
        try:
            if self.config["bionetgenAnalysis"] and self.returnArray:
                ls2b.postAnalyzeFile(
                    self.config["outputFile"],
                    self.config["bionetgenAnalysis"],
                    self.returnArray.database,
                    replaceLocParams=self.config["replaceLocParams"],
                    obs_map_file=self.config["obs_map_file"],
                )
        except Exception as e:
            self.logger.warning(
                "Post-analysis failed", loc=f"{__file__} : AtomizeTool.run()"
            )
            print("Post analysis failed")
            print(e)

        self.logger.debug(
            "Writing annotation file", loc=f"{__file__} : AtomizeTool.run()"
        )

        try:
            if self.config["annotation"] and self.returnArray:
                with open(self.config["outputFile"] + ".yml", "w") as f:
                    f.write(
                        yaml.dump(self.returnArray.annotation, default_flow_style=False)
                    )
        except Exception as e:
            self.logger.warning(
                "Failed to write annotation file", loc=f"{__file__} : AtomizeTool.run()"
            )
            print(e)
        return self.returnArray



import sys
import string
from xml.dom import minidom, Node


def walk(parent, outFile, level, database):  # [1]
    for node in parent.childNodes:
        if node.nodeType == Node.ELEMENT_NODE:
            # Write out the element name.
            printLevel(outFile, level)
            # outFile.write('Element: %s\n' % node.nodeName)
            # Write out the attributes.
            attrs = node.attributes  # [2]
            for attrName in attrs.keys():
                attrNode = attrs.get(attrName)
                attrValue = attrNode.nodeValue
                printLevel(outFile, level + 2)
                if attrName == "rdf:resource":
                    database.append(attrValue)

            # outFile.write('Attribute -- Name: %s  Value: %s\n' % \
            #    (attrName, attrValue))
            # Walk over any text nodes in the current node.
            content = []  # [3]
            for child in node.childNodes:
                if child.nodeType == Node.TEXT_NODE:
                    content.append(child.nodeValue)
            if content:
                strContent = string.join(content)
                printLevel(outFile, level)
            # outFile.write('Content: "')
            # outFile.write(strContent)
            # outFile.write('"\n')
            # Walk the child nodes.
            walk(node, outFile, level + 1, database)


def printLevel(outFile, level):
    for idx in range(level):
        # outFile.write('    ')
        pass


# [5]
def parseAnnotation(XMLString):
    outFile = ""
    if XMLString == "":
        return []
    doc = minidom.parseString(XMLString)
    rootNode = doc.documentElement
    level = 0
    database = []
    walk(rootNode, outFile, level, database)
    return database


def main():
    args = sys.argv[1:]
    if len(args) != 1:
        print("usage: python test.py infile.xml")
        sys.exit(-1)
    print(parseAnnotation(args[0]))


if __name__ == "__main__":
    main()



import re, os
from bionetgen.main import BioNetGen
from bionetgen.network.blocks import (
    NetworkGroupBlock,
    NetworkParameterBlock,
    NetworkReactionBlock,
    NetworkSpeciesBlock,
    NetworkCompartmentBlock,
    NetworkFunctionBlock,
    NetworkEnergyPatternBlock,
    NetworkPopulationMapBlock,
)


# This allows access to the CLIs config setup
app = BioNetGen()
app.setup()
conf = app.config["bionetgen"]
def_bng_path = conf["bngpath"]


class BNGNetworkParser:
    """
    Parser object that deals with reading in the BNGL file and
    setting up the model object

    Usage: BNGParser(bngl_path)
           BNGParser(bngl_path, BNGPATH)

    Attributes
    ----------
    bngfile : BNGFile
        BNGFile object that's responsible for .bngl file manipulations

    Methods
    -------
    parse_model(model_file)
        parses the BNGL model at the given path and adds everything to a given model object
    parse_xml(xml_str)
        parses given xml string and adds everything to a given model object
    """

    def __init__(self, path) -> None:
        self.path = path
        self.network_name = os.path.splitext(os.path.basename(path))[0]
        with open(self.path, "r") as f:
            self.network_lines = f.readlines()

    def parse_network(self, network_obj) -> None:
        """
        Will determine the parser route eventually and call the right
        parser
        """
        # network name
        network_obj.network_name = self.network_name
        # find blocks
        pblock = [-1, -1]
        sblock = [-1, -1]
        rblock = [-1, -1]
        gblock = [-1, -1]
        for iline, line in enumerate(self.network_lines):
            if line.strip() == "begin parameters":
                pblock[0] = iline
            if line.strip() == "end parameters":
                pblock[1] = iline
            if line.strip() == "begin species":
                sblock[0] = iline
            if line.strip() == "end species":
                sblock[1] = iline
            if line.strip() == "begin reactions":
                rblock[0] = iline
            if line.strip() == "end reactions":
                rblock[1] = iline
            if line.strip() == "begin groups":
                gblock[0] = iline
            if line.strip() == "end groups":
                gblock[1] = iline
        # add parameters
        if pblock[0] > 0 and pblock[1] > 0:
            param_block = NetworkParameterBlock()
            for iline in range(pblock[0] + 1, pblock[1]):
                m = re.match("([^#]*)(#.*)?", self.network_lines[iline])
                if m.group(1).strip() != "":
                    splt = m.group(1).split()
                    pid = splt[0]
                    pname = splt[1]
                    pvalue = splt[2]
                    comment = m.group(2)
                    param_block.add_parameter(pid, pname, pvalue, comment=comment)
            network_obj.add_block(param_block)
        # add species
        if sblock[0] > 0 and sblock[1] > 0:
            spec_block = NetworkSpeciesBlock()
            for iline in range(sblock[0] + 1, sblock[1]):
                m = re.match("([^#]*)(#.*)?", self.network_lines[iline])
                if m.group(1).strip() != "":
                    splt = m.group(1).split()
                    sid = splt[0]
                    name = splt[1]
                    try:
                        count = splt[2]
                    except:
                        import IPython

                        IPython.embed()
                    spec_block.add_species(sid, name, count)
            network_obj.add_block(spec_block)
        # add reactions
        if rblock[0] > 0 and rblock[1] > 0:
            rxns_block = NetworkReactionBlock()
            for iline in range(rblock[0] + 1, rblock[1]):
                m = re.match("([^#]*)(#.*)?", self.network_lines[iline])
                if m.group(1).strip() != "":
                    splt = m.group(1).split()
                    rid = splt[0]
                    reactants = splt[1].split(",")
                    products = splt[2].split(",")
                    rate_constant = splt[3]
                    comment = m.group(2)
                    rxns_block.add_reaction(
                        rid,
                        reactants=reactants,
                        products=products,
                        rate_constant=rate_constant,
                        comment=comment,
                    )
            network_obj.add_block(rxns_block)
        # # add groups
        if gblock[0] > 0 and gblock[1] > 0:
            grps_block = NetworkGroupBlock()
            for iline in range(gblock[0] + 1, gblock[1]):
                m = re.match("([^#]*)(#.*)?", self.network_lines[iline])
                if m.group(1).strip() != "":
                    splt = m.group(1).split()
                    rid = splt[0]
                    name = splt[1]
                    members = splt[2].split(",")
                    comment = m.group(2)
                    grps_block.add_group(rid, name, members, comment=comment)
            network_obj.add_block(grps_block)
        # import IPython,sys;IPython.embed();sys.exit()



from utils import readBNGXML
import argparse
from utils import extractAtomic
from collections import defaultdict, Counter


import collections


class OrderedSet(collections.MutableSet):
    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]  # sentinel node for doubly linked list
        self.map = {}  # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev, next = self.map.pop(key)
            prev[2] = next
            next[1] = prev

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError("set is empty")
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return "%s()" % (self.__class__.__name__,)
        return "%s(%r)" % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)


def extractCenterContext(rules, excludeReverse=False):
    transformationCenter = []
    transformationContext = []
    transformationProduct = []
    atomicArray = []
    actionNames = []
    label = []
    doubleModificationRules = []
    for idx, rule in enumerate(rules):
        (
            tatomicArray,
            ttransformationCenter,
            ttransformationContext,
            tproductElements,
            tactionNames,
            tlabelArray,
        ) = extractAtomic.extractTransformations([rule], True)
        if excludeReverse and "_reverse_" in rule[0].label:
            continue
        label.append(rule[0].label)

        if len([x for x in tactionNames if "ChangeCompartment" not in x]) > 1:
            doubleModificationRules.append(rule[0].label)
        transformationCenter.append(ttransformationCenter)
        transformationContext.append(ttransformationContext)

        actionNames.append(tactionNames)
        atomicArray.append(tatomicArray)
        transformationProduct.append(tproductElements)
    return (
        label,
        transformationCenter,
        transformationContext,
        transformationProduct,
        atomicArray,
        actionNames,
        doubleModificationRules,
    )


def getStateTransitionDiagram(
    labels, centers, products, contexts, actions, molecules, rules, parameters
):
    def isActive(state):
        return "!" in state or ("~" in state and "~0" not in state)

    moleculeDict = defaultdict(list)

    for molecule in molecules:
        for component in molecule.components:
            moleculeDict[molecule.name].append(component.name)
    nodeList = defaultdict(set)
    edgeList = defaultdict(OrderedSet)
    for label, center, product, context, action, rule in zip(
        labels, centers, products, contexts, actions, rules
    ):
        sourceCounter = {}
        destinationCounter = {}

        tmpSourceCounter = defaultdict(Counter)
        tmpContext = {}
        flag = True

        for centerUnit, productUnit, actionUnit, contextUnit in zip(
            center, product, action, context
        ):
            # create a node label based on reactant + context/ product + context
            if "ChangeCompartment" in actionUnit:
                continue
            if not flag:
                for species in centerUnit:
                    for element in species.split("."):
                        if element.split("(")[0].split("%")[0] not in sourceCounter:
                            sourceCounter[
                                element.split("(")[0].split("%")[0]
                            ] = Counter()
                            for component in moleculeDict[
                                element.split("(")[0].split("%")[0]
                            ]:
                                sourceCounter[element.split("(")[0].split("%")[0]][
                                    component
                                ] = 0
                        if isActive(element.split("(")[1][:-1]):
                            componentName = (
                                element.split("(")[1][:-1].split("~")[0].split("!")[0]
                            )
                            tmpSourceCounter[element.split("(")[0].split("%")[0]][
                                componentName
                            ] += centerUnit[species]
            else:
                for species in centerUnit:
                    for element in species.split("."):
                        if element.split("(")[0].split("%")[0] not in sourceCounter:
                            sourceCounter[
                                element.split("(")[0].split("%")[0]
                            ] = Counter()
                            for component in moleculeDict[
                                element.split("(")[0].split("%")[0]
                            ]:
                                sourceCounter[element.split("(")[0].split("%")[0]][
                                    component
                                ] = 0
                        if isActive(element.split("(")[1][:-1]):
                            componentName = (
                                element.split("(")[1][:-1].split("~")[0].split("!")[0]
                            )
                            sourceCounter[element.split("(")[0].split("%")[0]][
                                componentName
                            ] += centerUnit[species]
                    tmpContext[species] = contextUnit
                flag = False
            for species in productUnit:
                for element in species.split("."):
                    if element.split("(")[0].split("%")[0] not in destinationCounter:
                        destinationCounter[
                            element.split("(")[0].split("%")[0]
                        ] = Counter()
                        for component in moleculeDict[
                            element.split("(")[0].split("%")[0]
                        ]:
                            destinationCounter[element.split("(")[0].split("%")[0]][
                                component
                            ] = 0

                    if isActive(element.split("(")[1][:-1]):
                        componentName = (
                            element.split("(")[1][:-1].split("~")[0].split("!")[0]
                        )
                        destinationCounter[element.split("(")[0].split("%")[0]][
                            componentName
                        ] += productUnit[species]
        # add the first context unit
        if len(tmpContext) > 0:
            finalContext = tmpContext[tmpContext.keys()[0]]
            for idx in range(1, len(tmpContext.keys())):
                finalContext[tmpContext.keys()[idx]] -= 1
        else:
            finalContext = {}
        for species in finalContext:
            # for speciesUnit in tmpContext:
            #    for species in tmpContext[speciesUnit]:
            for element in species.split("."):
                if isActive(element.split("(")[1][:-1]):
                    if element.split("(")[0].split("%")[0] in sourceCounter:
                        componentName = (
                            element.split("(")[1][:-1].split("~")[0].split("!")[0]
                        )

                        sourceCounter[element.split("(")[0].split("%")[0]][
                            componentName
                        ] += finalContext[species]
                        destinationCounter[element.split("(")[0].split("%")[0]][
                            componentName
                        ] += finalContext[species]

        for element in tmpSourceCounter:
            destinationCounter[element].subtract(tmpSourceCounter[element])
        for molecule in sourceCounter:

            if molecule in destinationCounter:
                localMoleculeCounter = Counter(moleculeDict[molecule])
                # get total list of nodes. this is important because of symmetric components

                # sourceTuple = tuple(sorted([(x[0], x[1] > 0) for x in sourceCounter[molecule].items()], key=lambda x: x[0]))
                # destinationTuple = tuple(sorted([(x[0], x[1] > 0) for x in destinationCounter[molecule].items()], key=lambda x: x[0]))

                sourceTuple = []
                for x in sourceCounter[molecule].items():
                    sourceTuple.append((x[0], x[1] > 0))
                    for idx in range(1, localMoleculeCounter[x[0]]):
                        sourceTuple.append(
                            ("{0}-{1}".format(x[0], idx + 1), x[1] > idx)
                        )
                sourceTuple = tuple(sorted(sourceTuple, key=lambda x: x[0]))

                destinationTuple = []
                for x in destinationCounter[molecule].items():
                    destinationTuple.append((x[0], x[1] > 0))
                    for idx in range(1, localMoleculeCounter[x[0]]):
                        destinationTuple.append(
                            ("{0}-{1}".format(x[0], idx + 1), x[1] > idx)
                        )
                destinationTuple = tuple(sorted(destinationTuple, key=lambda x: x[0]))

                # sourceTuple = tuple(sorted([x[0] for x in sourceCounter[molecule].items()]))
                # destinationTuple = tuple(sorted([x[0] for x in destinationCounter[molecule].items() if x[1]> 0]))

                nodeList[molecule].add(sourceTuple)
                nodeList[molecule].add(destinationTuple)
                parameterList = [
                    parameters[x] if x in parameters else x for x in rule[0].rates
                ]
                edgeList[molecule].add(
                    (sourceTuple, destinationTuple, label, ",".join(parameterList))
                )
        # find the intersection context set
    return nodeList, edgeList


def getContextRequirements(
    inputfile, collapse=True, motifFlag=False, excludeReverse=False
):
    """
    Receives a BNG-XML file and returns the contextual dependencies implied by this file
    """
    molecules, rules, parameters = readBNGXML.parseXML(inputfile)

    moleculeStateMatrix = {}
    (
        label,
        center,
        context,
        product,
        atomicArray,
        actions,
        doubleAction,
    ) = extractCenterContext(rules, excludeReverse=excludeReverse)
    return getStateTransitionDiagram(
        label, center, product, context, actions, molecules, rules, parameters
    )


def getContextRequirementsFromNamespace(namespace, excludeReverse=False):
    (
        label,
        center,
        context,
        product,
        atomicArray,
        actions,
        doubleAction,
    ) = extractCenterContext(namespace["rules"], excludeReverse=excludeReverse)
    return getStateTransitionDiagram(
        label,
        center,
        product,
        context,
        actions,
        namespace["molecules"],
        namespace["rules"],
        namespace["parameters"],
    )


def defineConsole():
    """
    defines the program console line commands
    """
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument("-i", "--input", type=str, help="settings file", required=True)
    return parser


if __name__ == "__main__":
    parser = defineConsole()
    namespace = parser.parse_args()
    inputFile = namespace.input

    stateDictionary = getContextRequirements(inputFile, collapse=True, motifFlag=True)



# -*- coding: utf-8 -*-
"""
Created on Wed May 30 11:44:17 2012

@author: proto
"""
from copy import deepcopy
from lxml import etree
import re
from random import randint
from pyparsing import Word, Suppress, Optional, alphanums, Group, ZeroOrMore
from collections import Counter


def parseReactions(reaction):
    components = (
        Word(alphanums + "_")
        + Optional(Group("~" + Word(alphanums + "_")))
        + Optional(Group("!" + Word(alphanums + "+?")))
    )
    molecule = Word(alphanums + "_") + Optional(
        Suppress("(")
        + Group(components)
        + ZeroOrMore(Suppress(",") + Group(components))
        + Suppress(")")
    )

    species = Group(molecule) + ZeroOrMore(Suppress(".") + Group(molecule))

    result = species.parseString(reaction).asList()

    return result


def readFromString(string):
    sp = Species()
    reactionList = parseReactions(string)
    for molecule in reactionList:
        mol = Molecule(molecule[0], "")
        if len(molecule) > 1:
            for idx in range(1, len(molecule)):
                comp = Component(molecule[idx][0], "")
                if len(molecule[idx]) > 1:
                    for idx2 in range(1, len(molecule[idx])):
                        if molecule[idx][idx2][0] == "~":
                            comp.addState(molecule[idx][idx2][1])
                        elif molecule[idx][idx2][0] == "!":
                            comp.addBond(molecule[idx][idx2][1])
                mol.addComponent(comp)
        sp.addMolecule(mol)
    return sp


class Rule:
    def __init__(self, label=""):
        self.label = label
        self.reactants = []
        self.products = []
        self.rates = []
        self.bidirectional = False
        self.actions = []
        self.mapping = []

    def addReactant(self, reactant):
        self.reactants.append(reactant)

    def addProduct(self, product):
        self.products.append(product)

    def addReactantList(self, reactants):
        self.reactants.extend(reactants)

    def addProductList(self, products):
        self.products.extend(products)

    def addRate(self, rate):
        self.rates.append(rate)

    def addMapping(self, mapping):
        self.mapping.add(mapping)

    def addMappingList(self, mappingList):
        self.mapping.extend(mappingList)

    def addActionList(self, actionList):
        self.actions.extend(actionList)

    def __str__(self):
        finalStr = ""
        if self.label != "":
            finalStr += "{0}: ".format(self.label)
        arrow = " <-> " if self.bidirectional else " -> "
        finalStr += (
            " + ".join([str(x) for x in self.reactants])
            + arrow
            + " + ".join([str(x) for x in self.products])
            + " "
            + ",".join(self.rates)
        )
        return finalStr


class Species:
    def __init__(self):
        self.molecules = []
        self.bondNumbers = []
        self.bonds = []
        self.identifier = randint(0, 100000)
        self.idx = ""

    def getBondNumbers(self):
        bondNumbers = [0]
        for element in self.molecules:
            bondNumbers.extend(element.getBondNumbers())
        return bondNumbers

    def copy(self):
        species = Species()
        species.identifier = randint(0, 1000000)
        for molecule in self.molecules:
            species.molecules.append(molecule.copy())
        return species

    def getMoleculeById(self, idx):
        for molecule in self.molecules:
            if molecule.idx == idx:
                return molecule

    def addMolecule(self, molecule, concatenate=False, iteration=1):
        if not concatenate:
            self.molecules.append(molecule)
        else:
            counter = 1
            for element in self.molecules:

                if element.name == molecule.name:
                    if iteration == counter:
                        element.extend(molecule)
                        return
                    else:
                        counter += 1
            self.molecules.append(molecule)
            # self.molecules.append(molecule)
            # for element in self.molecules:
            #    if element.name == molecule.name:

    def addCompartment(self, tags):
        for molecule in self.molecules:
            molecule.setCompartment(tags)

    def deleteMolecule(self, moleculeName):
        deadMolecule = None
        for element in self.molecules:
            if element.name == moleculeName:
                deadMolecule = element
        if deadMolecule == None:
            return
        bondNumbers = deadMolecule.getBondNumbers()
        self.molecules.remove(deadMolecule)
        for element in self.molecules:
            for component in element.components:
                for number in bondNumbers:
                    if str(number) in component.bonds:
                        component.bonds.remove(str(number))

    def getMolecule(self, moleculeName):
        for molecule in self.molecules:
            if moleculeName == molecule.name:
                return molecule
        return None

    def getSize(self):
        return len(self.molecules)

    def getMoleculeNames(self):
        return [x.name for x in self.molecules]

    def contains(self, moleculeName):
        for molecule in self.molecules:
            if moleculeName == molecule.name:
                return True
        return False

    def addChunk(self, tags, moleculesComponents, precursors):
        """
        temporary transitional method
        """
        for (tag, components) in zip(tags, moleculesComponents):
            if self.contains(tag):
                tmp = self.getMolecule(tag)
            else:
                tmp = Molecule(tag)
                # for element in precursors:
                #    if element.getMolecule(tag) != None:
                #        tmp = element.getMolecule(tag)

            for component in components:
                if tmp.contains(component[0][0]):
                    tmpCompo = tmp.getComponent(component[0][0])
                    # continue
                else:
                    tmpCompo = Component(component[0][0])

                for index in range(1, len(component[0])):
                    tmpCompo.addState(component[0][index])
                if len(component) > 1:
                    tmpCompo.addBond(component[1])
                if not tmp.contains(component[0][0]):
                    tmp.addComponent(tmpCompo)
            if not self.contains(tag):
                self.molecules.append(tmp)

    def extend(self, species, update=True):
        if len(self.molecules) == len(species.molecules):
            for (selement, oelement) in zip(self.molecules, species.molecules):
                for component in oelement.components:
                    if component.name not in [x.name for x in selement.components]:
                        selement.components.append(component)
                    else:
                        for element in selement.components:
                            if element.name == component.name:
                                element.addStates(component.states, update)

        else:
            for element in species.molecules:
                if element.name not in [x.name for x in self.molecules]:

                    self.addMolecule(deepcopy(element), update)
                else:
                    for molecule in self.molecules:
                        if molecule.name == element.name:
                            for component in element.components:
                                if component.name not in [
                                    x.name for x in molecule.components
                                ]:
                                    molecule.addComponent(deepcopy(component), update)
                                else:
                                    comp = molecule.getComponent(component.name)
                                    for state in component.states:
                                        comp.addState(state, update)

    def updateBonds(self, bondNumbers):
        newBondNumbers = deepcopy(bondNumbers)
        correspondence = {}
        intersection = [int(x) for x in newBondNumbers if x in self.getBondNumbers()]
        for element in self.molecules:
            for component in element.components:
                for index in range(0, len(component.bonds)):
                    if int(component.bonds[index]) in intersection:

                        if component.bonds[index] in correspondence:
                            component.bonds[index] = correspondence[
                                component.bonds[index]
                            ]
                        else:
                            correspondence[component.bonds[index]] = (
                                max(intersection) + 1
                            )
                            component.bonds[index] = max(intersection) + 1
                        # intersection = [int(x) for x in newBondNumbers if x in self.getBondNumbers()]

    def append(self, species):
        newSpecies = deepcopy(species)
        newSpecies.updateBonds(self.getBondNumbers())

        for element in newSpecies.molecules:
            self.molecules.append(deepcopy(element))

    def sort(self):
        """
        Sort molecules by number of components, then number of bonded components, then the negative sum of the bond index, then number
        of active states, then string length
        """
        self.molecules.sort(
            key=lambda molecule: (
                len(molecule.components),
                -min(
                    [
                        int(y) if y not in ["?", "+"] else 999
                        for x in molecule.components
                        for y in x.bonds
                    ]
                    + [999]
                ),
                -len([x for x in molecule.components if len(x.bonds) > 0]),
                -len([x for x in molecule.components if x.activeState not in [0, "0"]]),
                len(str(molecule)),
                str(molecule),
            ),
            reverse=True,
        )

    def __str__(self):
        self.sort()
        name = ".".join([x.toString() for x in self.molecules])
        """
        name = name.replace('~','')
        
        name = name.replace('~','')
        name = name.replace(',','')
        name = name.replace('.','')
        name = name.replace('(','')
        name = name.replace(')','')
        name = name.replace('!','')
        name = name.replace('_','')
        """
        return name

    def str2(self):
        return ".".join([x.str2() for x in self.molecules])

    def reset(self):
        for element in self.molecules:
            element.reset()

    def toString(self):
        return self.__str__()

    def extractAtomicPatterns(self, action, site1, site2, differentiateDimers=False):
        atomicPatterns = {}
        bondedPatterns = {}
        reactionCenter = []
        context = []
        # one atomic pattern for the state, one for the bond
        nameCounter = Counter([x.name for x in self.molecules])
        nameCounterCopy = Counter([x.name for x in self.molecules])
        self.sort()
        for molecule in self.molecules:
            moleculeCounter = (
                nameCounter[molecule.name] - nameCounterCopy[molecule.name]
            )
            nameCounterCopy[molecule.name] -= 1
            for component in molecule.components:
                if component.activeState != "":
                    speciesStructure = Species()
                    # one atomic pattern for the states
                    speciesStructure.bonds = self.bonds
                    if differentiateDimers:
                        moleculeStructure = Molecule(
                            molecule.name + "%{0}".format(moleculeCounter), molecule.idx
                        )
                    else:
                        moleculeStructure = Molecule(molecule.name, molecule.idx)
                    componentStructure = Component(component.name, component.idx)

                    componentStructure.addState(component.activeState)
                    componentStructure.activeState = component.activeState
                    moleculeStructure.addComponent(componentStructure)
                    speciesStructure.addMolecule(moleculeStructure)
                    if (
                        componentStructure.idx in [site1, site2]
                        and action == "StateChange"
                    ):
                        reactionCenter.append((speciesStructure))
                    else:
                        context.append((speciesStructure))
                    atomicPatterns[str(speciesStructure)] = speciesStructure
                speciesStructure = Species()
                # one atomic pattern for the bonds
                speciesStructure.bonds = self.bonds
                if differentiateDimers:
                    moleculeStructure = Molecule(
                        molecule.name + "%{0}".format(moleculeCounter), molecule.idx
                    )
                else:
                    moleculeStructure = Molecule(molecule.name, molecule.idx)

                componentStructure = Component(component.name, component.idx)
                moleculeStructure.addComponent(componentStructure)
                speciesStructure.addMolecule(moleculeStructure)
                # atomicPatterns[str(speciesStructure)] = speciesStructure
                if len(component.bonds) == 0:
                    # if component.activeState == '':
                    atomicPatterns[str(speciesStructure)] = speciesStructure
                else:
                    if component.bonds[0] != "+":
                        componentStructure.addBond(1)
                    else:
                        componentStructure.addBond("+")
                    if component.bonds[0] not in bondedPatterns:
                        bondedPatterns[component.bonds[0]] = speciesStructure
                    elif (
                        "+" not in component.bonds[0]
                        or len(bondedPatterns[component.bonds[0]].molecules) == 0
                    ):
                        bondedPatterns[component.bonds[0]].addMolecule(
                            moleculeStructure
                        )
                if componentStructure.idx in [site1, site2] and action != "StateChange":
                    reactionCenter.append((speciesStructure))
                elif len(component.bonds) > 0 or component.activeState == "":
                    context.append((speciesStructure))
        for element in bondedPatterns:
            atomicPatterns[str(bondedPatterns[element])] = bondedPatterns[element]

        reactionCenter = [str(x) for x in reactionCenter if str(x) in atomicPatterns]
        context = [str(x) for x in context if str(x) in atomicPatterns]

        return atomicPatterns, reactionCenter, context

    def graphVizGraph(self, graph, identifier, layout="LR", options={}):
        speciesDictionary = {}
        graphName = "%s_%s" % (identifier, str(self))

        for idx, molecule in enumerate(self.molecules):
            ident = "%s_m%i" % (graphName, idx)
            speciesDictionary[molecule.idx] = ident
            if len(self.molecules) == 1:
                compDictionary = molecule.graphVizGraph(
                    graph, ident, flag=False, options=options
                )
            else:
                # s1 = graph.subgraph(name = graphName,label=' ')
                compDictionary = molecule.graphVizGraph(
                    graph, ident, flag=False, options=options
                )
            speciesDictionary.update(compDictionary)

        for bond in self.bonds:
            if bond[0] in speciesDictionary and bond[1] in speciesDictionary:
                if layout == "RL":
                    graph.add_edge(
                        speciesDictionary[bond[1]],
                        speciesDictionary[bond[0]],
                        dir="none",
                        len=0.1,
                        weight=100,
                    )
                else:
                    graph.add_edge(
                        speciesDictionary[bond[0]],
                        speciesDictionary[bond[1]],
                        dir="none",
                        len=0.1,
                        weight=100,
                    )
        return speciesDictionary

    def containsComponentIdx(self, idx, dictionary):
        for molecule in self.molecules:
            for component in molecule.components:
                if component.idx == idx:
                    return dictionary[idx]
        return None

    def notContainsComponentIdx(self, idx):
        context = []
        for molecule in self.molecules:
            for component in molecule.components:
                if component.idx not in idx:
                    context.append(component)
        return context

    def hasWildCardBonds(self):
        for molecule in self.molecules:
            if molecule.hasWildcardBonds():
                return True
        return False

    def listOfBonds(self, nameDict):
        listofbonds = {}
        for bond in self.bonds:
            mol1 = re.sub("_C[^_]*$", "", bond[0])
            mol2 = re.sub("_C[^_]*$", "", bond[1])
            if nameDict[mol1] not in listofbonds:
                listofbonds[nameDict[mol1]] = {}
            listofbonds[nameDict[mol1]][nameDict[bond[0]]] = [
                (nameDict[mol2], nameDict[bond[1]])
            ]
            if nameDict[mol2] not in listofbonds:
                listofbonds[nameDict[mol2]] = {}
            listofbonds[nameDict[mol2]][nameDict[bond[1]]] = [
                (nameDict[mol1], nameDict[bond[0]])
            ]

        return listofbonds


class Molecule:
    def __init__(self, name, idx):
        self.components = []
        self.name, self.idx = name, idx
        self.compartment = ""
        self.uniqueIdentifier = randint(0, 100000)

    def copy(self):
        molecule = Molecule(self.name, self.idx)
        for element in self.components:
            molecule.components.append(element.copy())
        return molecule

    def addChunk(self, chunk):
        component = Component(chunk[0][0][0][0])
        component.addState(chunk[0][0][0][1])
        self.addComponent(component)

    def addComponent(self, component, overlap=0):
        if not overlap:
            self.components.append(component)
        else:
            if not component.name in [x.name for x in self.components]:
                self.components.append(component)
            else:
                compo = self.getComponent(component.name)
                for state in component.states:
                    compo.addState(state)

    def setCompartment(self, compartment):
        self.compartment = compartment

    def getComponentById(self, idx):
        for component in self.components:
            if component.idx == idx:
                return component

    def getBondNumbers(self):
        bondNumbers = []
        for element in self.components:
            bondNumbers.extend([int(x) for x in element.bonds if x != "+"])
        return bondNumbers

    def getComponent(self, componentName):
        for component in self.components:
            if componentName == component.getName():
                return component

    def removeComponent(self, componentName):
        x = [x for x in self.components if x.name == componentName]
        if x != []:
            self.components.remove(x[0])

    def removeComponents(self, components):
        for element in components:
            if element in self.components:
                self.components.remove(element)

    def addBond(self, componentName, bondName):
        bondNumbers = self.getBondNumbers()
        while bondName in bondNumbers:
            bondName += 1
        component = self.getComponent(componentName)
        component.addBond(bondName)

    def getComponentWithBonds(self):
        return [x for x in self.components if x.bonds != []]

    def contains(self, componentName):
        return componentName in [x.name for x in self.components]

    def __str__(self):
        self.components = sorted(self.components, key=lambda st: st.name)
        finalStr = self.name
        if len(self.components) > 0:
            finalStr += "(" + ",".join([str(x) for x in self.components]) + ")"
        if self.compartment != "":
            finalStr += "@" + self.compartment
        return finalStr

    def toString(self):
        return self.__str__()

    def str2(self):
        self.components.sort()
        return self.name + "(" + ",".join([x.str2() for x in self.components]) + ")"

    def str3(self):
        return self.name + "(" + self.components[0].name + ")"

    def extend(self, molecule):
        for element in molecule.components:
            comp = [x for x in self.components if x.name == element.name]
            if len(comp) == 0:
                self.components.append(deepcopy(element))
            else:
                for bond in element.bonds:
                    comp[0].addBond(bond)
                for state in element.states:
                    comp[0].addState(state)

    def reset(self):
        for element in self.components:
            element.reset()

    def update(self, molecule):
        for comp in molecule.components:
            if comp.name not in [x.name for x in self.components]:
                self.components.append(deepcopy(comp))

    def graphVizGraph(self, graph, identifier, components=None, flag=False, options={}):
        moleculeDictionary = {}
        flag = False
        """
        if flag:
            graph.add_node(identifier,label=self.__str__(),name="cluster%s_%s" % (identifier,self.idx))
            print("cluster%s_%s" % (identifier,self.idx))
        else:
            print(identifier)
            graph.add_node(identifier,label=self.__str__(),name=identifier)
        
        moleculeDictionary[self.idx] = identifier
        return moleculeDictionary
        """
        if len(self.components) == 0:
            graph.add_node(identifier, label=self.name)
            moleculeDictionary[self.idx] = identifier
        else:
            if not flag:
                s1 = graph.subgraph(
                    name="cluster%s_%s" % (identifier, self.idx),
                    label=self.name,
                    **options
                )
            else:
                s1 = graph.subgraph(name=identifier, label=self.name, **options)
            s1.add_node(
                "cluster%s_%s_dummy" % (identifier, self.idx),
                shape="point",
                style="invis",
            )
            if components == None:
                tmpComponents = self.components
            else:
                tmpComponents = components
            for idx, component in enumerate(tmpComponents):
                ident = "%s_c%i" % (identifier, idx)
                moleculeDictionary[component.idx] = ident
                compDictionary = component.graphVizGraph(s1, ident)
                moleculeDictionary.update(compDictionary)
        return moleculeDictionary

    def hasWildcardBonds(self):
        for component in self.components:
            if component.hasWilcardBonds():
                return True
        return False

    def distance(self, cMolecule):
        distance = 0
        distance += 10000 if self.name != cMolecule.name else 0
        for component1, component2 in zip(self.components, cMolecule.components):
            distance += not component1.bonds == component2.bonds
            distance += not component1.activeState == component2.activeState
        return distance

    def compare(self, cMolecule):
        self.components = sorted(self.components, key=lambda st: st.name)
        cMolecule.components = sorted(cMolecule.components, key=lambda st: st.name)
        for c1, c2 in zip(self.components, cMolecule.components):
            if c1.activeState != c2.activeState:
                c1.activeState = ""

            if c1.bonds != c2.bonds:
                """
                if len(c1.bonds) != len(c2.bonds) or '?' in c1.bonds or '?' in c2.bonds:
                    c1.bonds = ['?']

                else:
                    c1.bonds = ['+']
                """


class Component:
    def __init__(self, name, idx, bonds=[], states=[]):
        self.name, self.idx = name, idx
        self.states = []
        self.bonds = []
        self.activeState = ""

    def copy(self):
        component = Component(
            self.name, self.idx, deepcopy(self.bonds), deepcopy(self.states)
        )
        component.activeState = deepcopy(self.activeState)
        return component

    def addState(self, state, update=True):
        if not state in self.states:
            self.states.append(state)
        if update:
            self.setActiveState(state)
        # print('LALALA',state)

    def addStates(self, states, update=True):
        for state in states:
            if state not in self.states:
                self.addState(state, update)

    def addBond(self, bondName):
        if not bondName in self.bonds:
            self.bonds.append(bondName)

    def setActiveState(self, state):
        if state not in self.states:
            return False
        self.activeState = state
        return True

    def getRuleStr(self):
        tmp = self.name
        if len(self.bonds) > 0:
            tmp += "!" + "!".join([str(x) for x in self.bonds])
        if self.activeState != "":
            tmp += "~" + self.activeState
        return tmp

    def getTotalStr(self):
        return self.name + "~".join(self.states)

    def getName(self):
        return self.name

    def __str__(self):
        return self.getRuleStr()

    def str2(self):
        tmp = self.name
        if len(self.bonds) > 0:
            tmp += "!" + "!".join([str(x) for x in self.bonds])
        if len(self.states) > 0:
            tmp += "~" + "~".join([str(x) for x in self.states])
        return tmp

    def __hash__(self):
        return self.name

    def reset(self):
        self.bonds = []
        if "0" in self.states:
            self.activeState = "0"

    def hasWilcardBonds(self):
        if "+" in self.bonds:
            return True
        return False

    def graphVizGraph(self, graph, identifier):
        compDictionary = {}

        if len(self.states) == 0:
            graph.add_node(identifier, label=self.name)
        else:
            s1 = graph.subgraph(
                name="cluster%s_%s" % (identifier, self.idx), label=self.name
            )
            if self.activeState != "":
                ident = "%s" % (identifier)
                compDictionary[self.activeState] = ident
                s1.add_node(ident, label=self.activeState)
            else:
                for idx, state in enumerate(self.state):
                    ident = "%s_s%i" % (identifier, idx)
                    s1.add_node(ident, label=state)
                    compDictionary[state] = ident
        return compDictionary

    def createGraph(self, identifier):

        return component, compDictionary


class States:
    def __init__(self, name="", idx=""):
        self.name = name
        self.idx = idx


class Action:
    def __init__(self):
        self.action = ""
        self.site1 = ""
        self.site2 = ""

    def setAction(self, action, site1, site2=""):
        self.action = action
        self.site1 = site1
        self.site2 = site2

    def __str__(self):
        return "%s, %s, %s" % (self.action, self.site1, self.site2)


class Databases:
    def __init__(self):
        self.translator = {}
        self.synthesisDatabase = {}
        self.catalysisDatabase = {}
        self.rawDatabase = {}
        self.labelDictionary = {}
        self.synthesisDatabase2 = {}

    def getRawDatabase(self):
        return self.rawDatabase

    def getLabelDictionary(self):
        return self.labelDictionary

    def add2LabelDictionary(self, key, value):
        temp = tuple(key)
        temp = temp.sort()
        self.labelDictionary[temp] = value

    def add2RawDatabase(self, rawDatabase):
        pass

    def getTranslator(self):
        return self.translator


if __name__ == "__main__":
    sp = readFromString("A(b!1,p~P).B(a!1)")



import os
from bionetgen.core.exc import BNGPerlError
import subprocess
from distutils import spawn

from bionetgen.core.utils.logging import BNGLogger


class ActionList:
    """
    Class to store everything related to BioNetGen actions.
    This class stores information about the list of BNG actions, their
    arguments as well as their syntax information. The class also provides
    an argument parser using pyparsing.

    Usage: ActionList()

    Attributes
    ----------
    normal_types : list

    no_setter_syntax : list
        actions without => syntax
    square_braces : list
        actions that use [] syntax in them
    before_model : list
        actions that are supposed to come before the model
    possible_types : list
        the list of all possible actions
    arg_dict : dict
        dictionary that contains every argument of every action
    irregular_args : dict
        actions that have arguments that aren't simply arg=>val

    Methods
    ---------
    is_before_model : bool
        checks if a given action is supposed to come before `begin model`
    define_parser : None
        sets ActionList.action_parser to a pyparsing parser that's capable of
        splitting up any BNG action into parts
    """

    def __init__(self):
        # these are all the action types, categorized
        # by their argument syntax
        self.normal_types = [
            "generate_network",
            "generate_hybrid_model",
            "simulate",
            "simulate_ode",
            "simulate_ssa",
            "simulate_pla",
            "simulate_nf",
            "parameter_scan",
            "bifurcate",
            "readFile",
            "writeFile",
            "writeModel",
            "writeNetwork",
            "writeXML",
            "writeSBML",
            "writeMfile",
            "writeCPYfile",
            "writeMexfile",
            "writeMDL",
            "visualize",
        ]
        self.no_setter_syntax = [
            "setConcentration",
            "addConcentration",
            "setParameter",
            "quit",
            "setModelName",
            "substanceUnits",
            "version",
            "setOption",
        ]
        self.square_braces = [
            "saveConcentrations",
            "resetConcentrations",
            "resetParameters",
            "saveParameters",
        ]
        # remember what's written before models
        self.before_model = [
            "setModelName",
            "substanceUnits",
            "version",
            "setOption",
        ]
        self.possible_types = (
            self.normal_types + self.no_setter_syntax + self.square_braces
        )
        # Use dictionary to keep track of all possible args (and types?) for each action
        self.arg_dict = {}
        # arg_dict["action"] = ["arg1", "arg2", "etc."]
        # normal_types
        self.arg_dict["generate_network"] = [
            "prefix",
            "suffix",
            "verbose",
            "overwrite",
            "print_iter",
            "max_agg",
            "max_iter",
            "max_stoich",
            "TextReaction",
            "TextSpecies",
        ]
        self.arg_dict["generate_hybrid_model"] = [
            "prefix",
            "suffix",
            "verbose",
            "overwrite",
            "actions",
            "execute",
            "safe",
        ]
        self.arg_dict["simulate"] = [
            "prefix",
            "suffix",
            "verbose",
            "method",
            "argfile",
            "continue",
            "t_start",
            "t_end",
            "n_steps",
            "n_output_steps",
            "sample_times",
            "output_step_interval",
            "max_sim_steps",
            "stop_if",
            "print_on_stop",
            "print_end",
            "print_net",
            "save_progress",
            "print_CDAT",
            "print_functions",
            "netfile",
            "seed",
            # TODO: arguments for a method called "psa" that is not documented in
            # https://docs.google.com/spreadsheets/d/1Co0bPgMmOyAFxbYnGCmwKzoEsY2aUCMtJXQNpQCEUag/
            "poplevel",
            "check_product_scale",
        ]
        self.arg_dict["simulate_ode"] = [
            "prefix",
            "suffix",
            "verbose",
            "argfile",
            "continue",
            "t_start",
            "t_end",
            "n_steps",
            "n_output_steps",
            "sample_times",
            "output_step_interval",
            "max_sim_steps",
            "stop_if",
            "print_on_stop",
            "print_end",
            "print_net",
            "save_progress",
            "print_CDAT",
            "print_functions",
            "netfile",
            "seed",
            "atol",
            "rtol",
            "sparse",
            "steady_state",
        ]
        self.arg_dict["simulate_ssa"] = [
            "prefix",
            "suffix",
            "verbose",
            "argfile",
            "continue",
            "t_start",
            "t_end",
            "n_steps",
            "n_output_steps",
            "sample_times",
            "output_step_interval",
            "max_sim_steps",
            "stop_if",
            "print_on_stop",
            "print_end",
            "print_net",
            "save_progress",
            "print_CDAT",
            "print_functions",
            "netfile",
            "seed",
        ]
        self.arg_dict["simulate_pla"] = [
            "prefix",
            "suffix",
            "verbose",
            "argfile",
            "continue",
            "t_start",
            "t_end",
            "n_steps",
            "n_output_steps",
            "sample_times",
            "output_step_interval",
            "max_sim_steps",
            "stop_if",
            "print_on_stop",
            "print_end",
            "print_net",
            "save_progress",
            "print_CDAT",
            "print_functions",
            "netfile",
            "seed",
            "pla_config",
            "pla_output",
        ]
        self.arg_dict["simulate_nf"] = [
            "prefix",
            "suffix",
            "verbose",
            "argfile",
            "continue",
            "t_start",
            "t_end",
            "n_steps",
            "n_output_steps",
            "sample_times",
            "output_step_interval",
            "max_sim_steps",
            "stop_if",
            "print_on_stop",
            "print_end",
            "print_net",
            "save_progress",
            "print_CDAT",
            "print_functions",
            "netfile",
            "seed",
            "complex",
            "nocslf",
            "notf",
            "binary_output",
            "gml",
            "equil",
            "get_final_state",
            "utl",
            "param",
        ]
        self.arg_dict["simulate"] = list(
            set(
                self.arg_dict["simulate"]
                + self.arg_dict["simulate_ode"]
                + self.arg_dict["simulate_ssa"]
                + self.arg_dict["simulate_pla"]
                + self.arg_dict["simulate_nf"]
            )
        )
        self.arg_dict["parameter_scan"] = [
            "prefix",
            "suffix",
            "verbose",
            "method",
            "argfile",
            "continue",
            "t_start",
            "t_end",
            "n_steps",
            "n_output_steps",
            "sample_times",
            "output_step_interval",
            "max_sim_steps",
            "stop_if",
            "print_on_stop",
            "print_end",
            "print_net",
            "save_progress",
            "print_CDAT",
            "print_functions",
            "netfile",
            "seed",
            "parameter",
            "par_min",
            "par_max",
            "n_scan_pts",
            "log_scale",
            "par_scan_vals",
            "reset_conc",
        ]
        self.arg_dict["parameter_scan"] = list(
            set(self.arg_dict["parameter_scan"] + self.arg_dict["simulate"])
        )
        self.arg_dict["bifurcate"] = [
            "prefix",
            "suffix",
            "verbose",
            "method",
            "argfile",
            "continue",
            "t_start",
            "t_end",
            "n_steps",
            "n_output_steps",
            "sample_times",
            "output_step_interval",
            "max_sim_steps",
            "stop_if",
            "print_on_stop",
            "print_end",
            "print_net",
            "save_progress",
            "print_CDAT",
            "print_functions",
            "netfile",
            "seed",
            "parameter",
            "par_min",
            "par_max",
            "n_scan_pts",
            "log_scale",
            "par_scan_vals",
        ]
        self.arg_dict["bifurcate"] = list(
            set(self.arg_dict["bifurcate"] + self.arg_dict["parameter_scan"])
        )
        self.arg_dict["bifurcate"].remove("reset_conc")
        self.arg_dict["readFile"] = ["file", "blocks", "atomize", "skip_actions"]
        self.arg_dict["writeFile"] = [
            "format",
            "prefix",
            "suffix",
            "evaluate_expressions",
            "include_model",
            "include_network",
            "overwrite",
            "pretty_formatting",
            "TextReaction",
            "TextSpecies",
        ]
        self.arg_dict["writeModel"] = [
            "format",
            "prefix",
            "suffix",
            "evaluate_expressions",
            "include_model",
            "include_network",
            "overwrite",
            "pretty_formatting",
            "TextReaction",
            "TextSpecies",
        ]
        self.arg_dict["writeNetwork"] = [
            "format",
            "prefix",
            "suffix",
            "evaluate_expressions",
            "include_model",
            "include_network",
            "overwrite",
            "pretty_formatting",
            "TextReaction",
            "TextSpecies",
        ]
        self.arg_dict["writeXML"] = [
            "format",
            "prefix",
            "suffix",
            "evaluate_expressions",
            "include_model",
            "include_network",
            "overwrite",
            "pretty_formatting",
            "TextReaction",
            "TextSpecies",
        ]
        self.arg_dict["writeSBML"] = ["prefix", "suffix"]
        self.arg_dict["writeMfile"] = [
            "prefix",
            "suffix",
            "t_start",
            "t_end",
            "n_steps",
            "atol",
            "rtol",
            "max_step",
            "bdf",
            "maxOrder",
            "stats",
        ]
        self.arg_dict["writeCPYfile"] = [
            "prefix",
            "suffix",
            "t_start",
            "t_end",
            "n_steps",
            "atol",
            "rtol",
            "max_step",
            "bdf",
            "maxOrder",
            "stats",
        ]
        self.arg_dict["writeMexfile"] = [
            "prefix",
            "suffix",
            "t_start",
            "t_end",
            "n_steps",
            "atol",
            "rtol",
            "max_step",
            "max_num_steps",
            "max_err_test_fails",
            "max_conv_fails",
            "stiff",
            "sparse",
        ]
        self.arg_dict["writeMDL"] = ["prefix", "suffix"]
        self.arg_dict["visualize"] = [
            "type",
            "help",
            "suffix",
            "each",
            "background",
            "groups",
            "collapse",
            "filter",
            "level",
            "textonly",
            "opts",
        ]
        # no_setter_syntax
        self.arg_dict["setConcentration"] = []
        self.arg_dict["addConcentration"] = []
        self.arg_dict["setParameter"] = []
        self.arg_dict["saveParameters"] = []
        self.arg_dict["quit"] = None
        self.arg_dict["setModelName"] = []
        self.arg_dict["substanceUnits"] = []
        self.arg_dict["version"] = []
        self.arg_dict["setOption"] = []
        # square_braces
        self.arg_dict["saveConcentrations"] = []
        self.arg_dict["resetConcentrations"] = []
        self.arg_dict["resetParameters"] = []

        # irregular arg types
        self.irregular_args = {}
        self.irregular_args["max_stoich"] = "dict"
        self.irregular_args["actions"] = "list"
        self.irregular_args["sample_times"] = "list"
        self.irregular_args["par_scan_vals"] = "list"
        self.irregular_args["blocks"] = "list"
        self.irregular_args["opts"] = "list"

    def is_before_model(self, action_name):
        if action_name in self.before_model:
            return True
        return False

    def define_parser(self):
        ## Define action grammar
        import pyparsing as pp

        #
        base_name = pp.Word(pp.alphas, pp.alphanums + "_")
        action_name = base_name
        #
        dquote_word = pp.dblQuotedString
        squote_word = pp.sglQuotedString
        quote_word = dquote_word ^ squote_word
        # all action argument types
        # TODO: deal w/ zero argument list
        list_arg = "[" + pp.delimitedList(quote_word) + "]"
        #
        arg_type_bool = pp.Word("0") ^ pp.Word("1")
        arg_type_int = pp.Word(pp.nums)
        arg_type_float = pp.Word(pp.nums + ".")
        arg_type_expr = pp.Word(
            pp.nums + "." + "+" + "-" + "e" + "E" + "(" + ")" + "/" + "*" + "^"
        )
        arg_type_list = "[" + pp.delimitedList((quote_word ^ arg_type_float)) + "]"
        arg_type_string = quote_word
        #
        curly_arg_token = quote_word + "=>" + arg_type_int
        # TODO: handle 0 case
        arg_type_curly = "{" + pp.delimitedList(curly_arg_token) + "}"
        arg_types = (
            arg_type_bool
            ^ arg_type_int
            ^ arg_type_float
            ^ arg_type_list
            ^ arg_type_list
            ^ arg_type_string
            ^ arg_type_curly
            ^ arg_type_expr
        )
        #
        one_arg = quote_word
        two_arg = quote_word + "," + (arg_type_expr ^ quote_word)
        #
        single_arg = base_name + "=>" + arg_types
        #
        reg_arg_full = "{" + pp.Optional(pp.delimitedList(single_arg)) + "}"
        #
        reg_action_tk = (
            action_name + "(" + reg_arg_full + ")" + pp.Optional(";") + pp.stringEnd
        )
        two_arg_action_tk = (
            action_name
            + "("
            + quote_word
            + ","
            + pp.SkipTo(")" + pp.Optional(";") + pp.stringEnd)
            + ")"
            + pp.Optional(";")
            + pp.stringEnd
        )
        one_arg_action_tk = (
            action_name
            + "("
            + pp.Optional(one_arg)
            + ")"
            + pp.Optional(";")
            + pp.stringEnd
        )
        list_arg_action_tk = (
            action_name + "(" + list_arg + ")" + pp.Optional(";") + pp.stringEnd
        )
        full_action_tk = (
            reg_action_tk ^ list_arg_action_tk ^ two_arg_action_tk ^ one_arg_action_tk
        )
        ## Action grammar done
        self.action_parser = full_action_tk


def find_BNG_path(BNGPATH=None):
    """
    A simple function finds the path to BNG2.pl from
    * Environment variable
    * Assuming it's under PATH
    * Given optional path as argument

    Usage: test_bngexec(path)
           test_bngexec()

    Arguments
    ---------
    BNGPATH : str
        (optional) path to the folder that contains BNG2.pl
    """
    # TODO: Figure out how to use the BNG2.pl if it's set
    # in the PATH variable. Solution: set os.environ BNGPATH
    # and make everything use that route

    # Let's keep up the idea we pull this path from the environment
    if BNGPATH is None:
        try:
            BNGPATH = os.environ["BNGPATH"]
        except:
            pass
    # if still none, try pulling it from cmd line
    if BNGPATH is None:
        bngexec = "BNG2.pl"
        if test_bngexec(bngexec):
            # print("BNG2.pl seems to be working")
            # get the source of BNG2.pl
            BNGPATH = spawn.find_executable("BNG2.pl")
            BNGPATH, _ = os.path.split(BNGPATH)
    else:
        bngexec = os.path.join(BNGPATH, "BNG2.pl")
        if not test_bngexec(bngexec):
            RuntimeError("BNG2.pl is not working")
    return BNGPATH, bngexec


def test_perl(app=None, perl_path=None):
    """
    Test if perl is working

    Arguments
    ---------
    perl_path : str
        (optional) path to the folder that contains perl
    """
    logger = BNGLogger(app=app)
    logger.debug("Checking if perl is installed.", loc=f"{__file__} : test_perl()")
    # find path to perl binary
    if perl_path is None:
        perl_path = spawn.find_executable("perl")
    if perl_path is None:
        raise BNGPerlError
    # check if perl is actually working
    command = [perl_path, "-v"]
    rc, _ = run_command(command, suppress=True)
    if rc != 0:
        raise BNGPerlError


def test_bngexec(bngexec):
    """
    A simple function that test if BNG2.pl given runs

    Usage: test_bngexec(path)

    Arguments
    ---------
    bngexec : str
        path to BNG2.pl to test
    """
    command = ["perl", bngexec]
    rc, _ = run_command(command, suppress=True)
    if rc == 0:
        return True
    else:
        return False


def run_command(command, suppress=False, timeout=None):
    """
    A convenience function to run a given command. The command should be
    given as a list of values e.g. ['command', 'arg1', 'arg2'] etc.

    Suppress kwarg suppresses all output from the command and timeout kwarg
    allows you to set a time period in seconds after which the command will
    be killed.
    """
    if timeout is not None:
        # I am unsure how to do both timeout and the live polling of stdo
        rc = subprocess.run(command, timeout=timeout, capture_output=True)
        return rc.returncode, rc
    else:
        if suppress:
            process = subprocess.Popen(
                command,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                bufsize=-1,
            )
            rc = process.wait()
            return rc, process
        else:
            process = subprocess.Popen(command, stdout=subprocess.PIPE, encoding="utf8")
            out = []
            while True:
                output = process.stdout.readline()
                if output == "" and process.poll() is not None:
                    break
                if output:
                    o = output.strip()
                    out.append(o)
                    print(o)
            rc = process.wait()
            return rc, out



from .bngsimulator import BNGSimulator

# TODO: Unified API for all simulator objects here
class libRRSimulator(BNGSimulator):
    """
    libRoadRunner simulator wrapper


    Attributes
    ----------
    sbml: str
        the SBML used by the underlying libRoadRunner simulator

    Properties
    ----------
    simulator: obj
        the python object that runs the actual simulation.
        Any subclass implementing a BNGSimulator should set
        this property and make it so that setting the simulator
        object to the file path given to initialize the
        BNGSimulator class initializes the simulator object in turn.

    Methods
    -------
    simulate(args)
        Uses the arguments provided to call the underlying simulator
    """

    @property
    def simulator(self):
        """
        simulator attribute that stores
        the instantiated simulator object
        and also saves the SBML text in the
        sbml attribute
        """
        return self._simulator

    @simulator.setter
    def simulator(self, model):
        try:
            import roadrunner as rr

            self._simulator = rr.RoadRunner(model)
        except ImportError:
            print("libroadrunner is not installed!")

    @property
    def sbml(self):
        """
        sbml attribute which is just the SBML
        string with which the libRR instance
        is instantiated with
        """
        if not hasattr(self, "_sbml"):
            self._sbml = self.simulator.getCurrentSBML()
        return self._sbml

    @sbml.setter
    def sbml(self, model_str):
        self._sbml = model_str

    def simulate(self, *args, **kwargs):
        """
        generic simulate front-end that passes the
        args and kwargs to the underlying simulator object
        """
        return self.simulator.simulate(*args, **kwargs)
